# pylint: disable=no-self-use
import os

import numpy as np
import pytest
from skimage.morphology import diamond

from PartSegImage import Channel, ChannelInfo, Image, ImageWriter, TiffImageReader
from PartSegImage.image import FRAME_THICKNESS, _hex_to_rgb, _name_to_rgb


class TestImageBase:
    image_class = Image

    def needed_shape(self, shape, axes: str, drop: str):
        new_axes = self.image_class.array_axis_order
        return self._needed_shape(shape, axes, drop, new_axes)

    def needed_layer_shape(self, shape, axes: str, drop: str):
        new_axes = self.image_class.axis_order
        return self._needed_shape(shape, axes, drop, new_axes)

    def _needed_shape(self, shape, axes: str, drop: str, new_axes):
        for el in drop:
            new_axes = new_axes.replace(el, "")
            axes = axes.replace(el, "")
        res_shape = [1] * len(new_axes)
        for size, name in zip(shape, axes):
            res_shape[new_axes.index(name)] = size
        return tuple(res_shape)

    def image_shape(self, shape, axes):
        return self.needed_shape(shape, axes, "C")

    def mask_shape(self, shape, axes):
        return self.needed_shape(shape, axes, "C")

    def reorder_axes_letter(self, letters: str):
        res = "".join(x for x in self.image_class.axis_order if x in letters)
        assert len(res) == len(letters)
        return res

    def prepare_mask_shape(self, shape):
        base_axes = set("TZYX")
        refer_axes = self.image_class.axis_order.replace("C", "")
        i, j = 0, 0
        new_shape = [1] * len(refer_axes)
        for i, val in enumerate(refer_axes):
            if val in base_axes:
                new_shape[i] = shape[j]
                j += 1
        return new_shape

    def prepare_image_initial_shape(self, shape, channel):
        new_shape = self.prepare_mask_shape(shape)
        new_shape.insert(self.image_class.axis_order.index("C"), channel)
        return new_shape

    def test_fit_mask_simple(self):
        initial_shape = self.prepare_image_initial_shape([1, 10, 20, 20], 1)
        data = np.zeros(initial_shape, np.uint8)
        image = self.image_class(data, spacing=(1, 1, 1), file_path="", axes_order=self.image_class.axis_order)
        mask = np.zeros((1, 10, 20, 20), np.uint8)
        mask[0, 2:-2, 4:-4, 4:-4] = 5
        image.fit_mask_to_image(mask)

    def test_fit_mask_mapping_val(self):
        initial_shape = self.prepare_image_initial_shape([1, 10, 20, 20], 1)
        data = np.zeros(initial_shape, np.uint8)
        image = self.image_class(data, spacing=(1, 1, 1), file_path="", axes_order=self.image_class.axis_order)
        mask = np.zeros((1, 10, 20, 20), np.uint16)
        mask[0, 2:-2, 4:-4, 4:10] = 5
        mask[0, 2:-2, 4:-4, 11:-4] = 7
        mask2 = image.fit_mask_to_image(mask)
        assert np.all(np.unique(mask2) == [0, 1, 2])
        assert np.all(np.unique(mask) == [0, 5, 7])
        map_arr = np.array([0, 0, 0, 0, 0, 1, 0, 2])
        assert np.all(map_arr[mask] == mask2)
        assert mask2.dtype == np.uint8

    def test_fit_mask_to_image_change_type(self):
        initial_shape = self.prepare_image_initial_shape([1, 30, 50, 50], 1)
        data = np.zeros(initial_shape, np.uint8)
        image = self.image_class(data, spacing=(1, 1, 1), file_path="", axes_order=self.image_class.axis_order)
        mask_base = np.zeros(30 * 50 * 50, dtype=np.uint32)
        mask_base[:50] = np.arange(50, dtype=np.uint32)
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint8

        mask_base[:50] = np.arange(50, dtype=np.uint32) + 5
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint8

        mask_base[:350] = np.arange(350, dtype=np.uint32)
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint16

        mask_base[:350] = np.arange(350, dtype=np.uint32) + 5
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint16

        mask_base[: 2**16 + 5] = np.arange(2**16 + 5, dtype=np.uint32)
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint32

        mask_base[: 2**16 + 5] = np.arange(2**16 + 5, dtype=np.uint32) + 5
        image.set_mask(np.reshape(mask_base, (1, 30, 50, 50)))
        assert image.mask.dtype == np.uint32

    def test_image_mask(self):
        initial_shape = self.prepare_image_initial_shape([1, 10, 50, 50], 4)
        self.image_class(
            np.zeros(initial_shape),
            spacing=(5, 5, 5),
            mask=np.zeros((10, 50, 50)),
            axes_order=self.image_class.axis_order,
        )
        self.image_class(
            np.zeros(initial_shape),
            spacing=(5, 5, 5),
            mask=np.zeros((1, 10, 50, 50)),
            axes_order=self.image_class.axis_order,
        )
        mask = np.zeros((1, 10, 50, 50))
        mask[0, 2:-2] = 1
        im = self.image_class(
            np.zeros(initial_shape), spacing=(5, 5, 5), mask=mask, axes_order=self.image_class.axis_order
        )
        assert np.all(im.mask == mask)

    def test_image_mask_exception(self):
        if "C" not in self.image_class.axis_order:
            pytest.skip("Lack of channel axis")
        data_shape = (1,) * (len(self.image_class.axis_order) - 4) + (10, 50, 50, 4)
        with pytest.raises(ValueError, match="Wrong array shape"):
            self.image_class(
                np.zeros(data_shape),
                spacing=(5, 5, 5),
                mask=np.zeros(data_shape[:-2] + (40,)),
                axes_order=self.image_class.axis_order,
            )
        with pytest.raises(ValueError, match="Wrong array shape"):
            self.image_class(
                np.zeros(data_shape),
                spacing=(5, 5, 5),
                mask=np.zeros(data_shape),
                axes_order=self.image_class.axis_order,
            )

    def test_reorder_axes(self):
        fixed_array = self.image_class.reorder_axes(np.zeros((10, 20)), axes="XY")
        assert fixed_array.shape == self.image_shape((10, 20), "XY")
        fixed_image = self.image_class(np.zeros((10, 20)), spacing=(1, 1, 1), axes_order="XY")
        assert fixed_image.shape == self.image_shape((10, 20), "XY")
        with pytest.raises(ValueError, match="need to be equal to length of axes"):
            Image.reorder_axes(np.zeros((10, 20)), axes="XYZ")

    def test_reorder_axes_with_mask(self):
        im = self.image_class(
            np.zeros((10, 50, 50, 4)), spacing=(5, 5, 5), mask=np.zeros((10, 50, 50)), axes_order="ZYXC"
        )
        assert im.shape == self.image_shape((10, 50, 50, 4), "ZYXC")  # (1, 10, 50, 50, 4)
        im = self.image_class(
            np.zeros((50, 10, 50, 4)), spacing=(5, 5, 5), mask=np.zeros((50, 10, 50)), axes_order="YZXC"
        )
        assert im.shape == self.image_shape((50, 10, 50, 4), "YZXC")  # (1, 10, 50, 50, 4)

    def test_wrong_dim_create(self):
        with pytest.raises(ValueError, match="Data should"):
            self.image_class(np.zeros((10, 20)), spacing=(1, 1, 1), axes_order="XYZ")

        with pytest.raises(ValueError, match="Data should"):
            self.image_class([np.zeros((10, 20)), np.zeros((10,))], spacing=(1, 1, 1), axes_order="XYZ")

    def test_get_dimension_number(self):
        assert (
            self.image_class(
                np.zeros((1, 10, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 3
        )
        assert (
            self.image_class(
                np.zeros((1, 1, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 2
        )
        assert (
            self.image_class(
                np.zeros((10, 1, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 3
        )
        assert (
            self.image_class(
                np.zeros((1, 1, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 2
        )
        assert (
            self.image_class(
                np.zeros((10, 1, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 3
        )
        assert (
            self.image_class(
                np.zeros((10, 3, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
            ).get_dimension_number()
            == 4
        )

    def test_get_dimension_letters(self):
        assert self.image_class(
            np.zeros((1, 10, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("ZYX")
        assert self.image_class(
            np.zeros((1, 1, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("YX")
        assert self.image_class(
            np.zeros((10, 1, 20, 20, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("TYX")
        assert self.image_class(
            np.zeros((1, 1, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("YX")
        assert self.image_class(
            np.zeros((10, 1, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("TYX")
        assert self.image_class(
            np.zeros((10, 3, 20, 20, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        ).get_dimension_letters() == self.reorder_axes_letter("TZYX")

    def test_set_mask(self):
        initial_shape = self.prepare_image_initial_shape([1, 10, 20, 30], 1)
        image = self.image_class(
            np.zeros(initial_shape, np.uint8), spacing=(1, 1, 1), file_path="", axes_order=self.image_class.axis_order
        )
        assert image.mask is None
        assert not image.has_mask
        image.set_mask(np.ones((10, 20, 30), np.uint8))
        assert image.mask.shape == tuple(self.prepare_mask_shape((1, 10, 20, 30)))
        assert image.has_mask
        assert np.all(image.mask == 1)
        image.set_mask(np.full((10, 20, 30), 5, np.uint8))
        assert image.mask.shape == tuple(self.prepare_mask_shape((1, 10, 20, 30)))
        assert np.all(image.mask == 1)
        mask = np.full((10, 20, 30), 5, np.uint8)
        mask[0, 0, 0] = 1
        image.set_mask(mask)
        assert image.mask.shape == tuple(self.prepare_mask_shape((1, 10, 20, 30)))
        assert np.all(np.bincount(image.mask.flat) == (0, 1, 10 * 20 * 30 - 1))
        image.set_mask(None)
        assert image.mask is None

    def test_set_mask_reorder(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 1), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        image.set_mask(np.ones((30, 20, 10), np.uint8), "XYZ")
        assert image.mask.shape == self.mask_shape((1, 10, 20, 30), "TZYX")

    def test_get_image_for_save(self):
        if "C" not in self.image_class.axis_order:
            pytest.skip("No channel axis")
        image = self.image_class(
            np.zeros((1, 10, 3, 20, 30), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZCYX"
        )
        assert image.get_image_for_save().shape == (1, 10, 3, 20, 30)
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        assert image.get_image_for_save().shape == (1, 10, 3, 20, 30)

    def test_get_image_for_save_no_channel(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYX"
        )
        assert image.get_image_for_save().shape == (1, 10, 1, 20, 30)
        image = self.image_class(
            np.zeros((1, 10, 20, 30), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYX"
        )
        assert image.get_image_for_save().shape == (1, 10, 1, 20, 30)

    def test_get_mask_for_save(self):
        image = self.image_class(
            np.zeros((1, 10, 3, 20, 30), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZCYX"
        )
        assert image.get_mask_for_save() is None
        image.set_mask(np.zeros((1, 10, 20, 30), np.uint8), axes="TZYX")
        assert image.get_mask_for_save().shape == (1, 10, 1, 20, 30)

    def test_image_properties(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        assert not image.has_mask
        assert not image.is_time
        assert image.is_stack
        assert not image.is_2d
        assert image.channels == 3
        assert image.layers == 10
        assert image.times == 1
        assert image.plane_shape == (20, 30)

    def test_swap_time_and_stack(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        image2 = image.swap_time_and_stack()
        assert image.times == 1
        assert image.layers == 10
        assert image2.times == 10
        assert image2.layers == 1

    def test_get_channel(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8),
            spacing=(1, 1, 1),
            file_path="",
            axes_order="TZYXC",
            channel_info=[ChannelInfo(name="a"), ChannelInfo(name="b"), ChannelInfo(name="c")],
        )
        assert image.has_channel(1)
        assert image.has_channel(Channel(1))
        assert not image.has_channel(5)
        assert not image.has_channel(Channel(5))
        channel = image.get_channel(1)
        assert channel.shape == self.mask_shape((1, 10, 20, 30), "TZYX")
        assert image.has_channel("b")
        assert image.has_channel(Channel("b"))
        assert not image.has_channel("d")
        assert not image.has_channel(Channel("d"))
        channel = image.get_channel(Channel("b"))
        assert channel.shape == self.mask_shape((1, 10, 20, 30), "TZYX")
        with pytest.raises(IndexError):
            image.get_channel(4)

    def test_get_layer(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        with pytest.deprecated_call():
            layer = image.get_layer(0, 5)
        assert layer.shape == self.needed_layer_shape((20, 30, 3), "YXC", "TZ")

    def test_spacing(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        assert image.spacing == (1, 1, 1)
        assert image.voxel_size == (1, 1, 1)
        image.set_spacing((1, 2, 3))
        assert image.spacing == (1, 2, 3)
        assert image.voxel_size == (1, 2, 3)
        with pytest.raises(ValueError, match="Correction of spacing fail"):
            image.set_spacing((1, 2, 3, 4))
        with pytest.raises(TypeError, match="is not iterable"):
            # noinspection PyTypeChecker
            image.set_spacing(1)
        image.set_spacing((1, 0, 4))
        assert image.spacing == (1, 2, 3)
        assert image.voxel_size == (1, 2, 3)

    def test_spacing_2d(self):
        image = self.image_class(
            np.zeros((1, 1, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        assert image.spacing == (1, 1)
        assert image.voxel_size == (1, 1)
        image.set_spacing((1, 2))
        assert image.spacing == (1, 2)
        assert image.voxel_size == (1, 2)

    def test_cut_image(self):
        # TODO add tests for more irregular shape
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(1, 1, 1), file_path="", axes_order="TZYXC"
        )
        mask1 = np.zeros((1, 10, 20, 30), np.uint8)
        mask1[0, 2:-2, 2:9, 2:-2] = 1
        mask1[0, 2:-2, 11:-2, 2:-2] = 2
        image.set_mask(mask1, "TZYX")
        mask = image.mask
        image.set_mask(np.zeros((1, 10, 20, 30), np.uint8), "TZYX")
        im = image.cut_image(mask == 1, replace_mask=False)
        assert np.all(im.mask == 0)
        assert im.shape == self.image_shape((1, 10, 11, 30, 3), axes="TZYXC")
        im = image.cut_image(mask == 2, replace_mask=True)
        assert np.all(
            im.mask[
                :, FRAME_THICKNESS:-FRAME_THICKNESS, FRAME_THICKNESS:-FRAME_THICKNESS, FRAME_THICKNESS:-FRAME_THICKNESS
            ]
            == 1
        )
        assert im.shape == self.image_shape((1, 10, 11, 30, 3), axes="TZYXC")

        # Test cutting with list of slices
        points = np.nonzero(mask == 2)
        lower_bound = np.min(points, axis=1)
        upper_bound = np.max(points, axis=1)
        cut_list = [slice(x, y + 1) for x, y in zip(lower_bound, upper_bound)]
        res = image.cut_image(cut_list)
        shape = [y - x + +1 for x, y in zip(lower_bound, upper_bound)]
        shape[image.x_pos] += 2 * FRAME_THICKNESS
        shape[image.y_pos] += 2 * FRAME_THICKNESS
        shape[image.stack_pos] += 2 * FRAME_THICKNESS
        assert res.shape == tuple(shape)

    def test_get_ranges(self):
        data = np.zeros((1, 10, 20, 30, 3), np.uint8)
        data[..., :10, 0] = 2
        data[..., :10, 1] = 20
        data[..., :10, 2] = 9
        image = self.image_class(data, spacing=(1, 1, 1), file_path="", axes_order="TZYXC")
        assert len(image.get_ranges()) == 3
        assert image.get_ranges() == [(0, 2), (0, 20), (0, 9)]

    def test_get_um_spacing(self):
        image = self.image_class(
            np.zeros((1, 10, 20, 30, 3), np.uint8), spacing=(10**-6, 10**-6, 10**-6), file_path="", axes_order="TZYXC"
        )
        assert image.get_um_spacing() == (1, 1, 1)
        image = self.image_class(
            np.zeros((1, 1, 20, 30, 3), np.uint8), spacing=(10**-6, 10**-6, 10**-6), file_path="", axes_order="TZYXC"
        )
        assert image.get_um_spacing() == (1, 1)

    def test_save(self, tmp_path):
        if "C" not in self.image_class.axis_order:
            pytest.skip("No channel axis")
        data = np.zeros((1, 10, 20, 30, 3), np.uint8)
        data[..., :10, 0] = 2
        data[..., :10, 1] = 20
        data[..., :10, 2] = 9
        image = self.image_class(data, spacing=(10**-6, 10**-6, 10**-6), file_path="", axes_order="TZYXC")
        mask = np.zeros((10, 20, 30), np.uint8)
        mask[..., 2:12] = 1
        image.set_mask(mask, "ZYX")
        ImageWriter.save(image, os.path.join(tmp_path, "img.tif"))
        ImageWriter.save_mask(image, os.path.join(tmp_path, "img_mask.tif"))
        read_image: Image = TiffImageReader.read_image(
            os.path.join(tmp_path, "img.tif"), os.path.join(tmp_path, "img_mask.tif")
        )
        assert read_image.get_um_spacing() == (1, 1, 1)
        assert len(read_image.get_ranges()) == 3
        assert read_image.get_ranges() == [(0, 2), (0, 20), (0, 9)]

    def test_axes_pos(self):
        data = np.zeros((10, 10), np.uint8)
        image = self.image_class(data, spacing=(1, 1), axes_order="XY")
        assert image.x_pos == image.array_axis_order.index("X")
        assert image.y_pos == image.array_axis_order.index("Y")
        assert image.time_pos == image.array_axis_order.index("T")
        assert image.stack_pos == image.array_axis_order.index("Z")

    def test_get_axis_positions(self):
        data = np.zeros((10, 10), np.uint8)
        image = self.image_class(data, spacing=(1, 1), axes_order="XY")
        assert set(image.get_axis_positions()) == set(image.axis_order)
        assert len(set(image.get_axis_positions().values())) == len(image.axis_order)

    def test_get_data(self):
        data = np.zeros((10, 10), np.uint8)
        image = self.image_class(data, spacing=(1, 1), axes_order="XY")
        assert np.all(image.get_data() == data)


class ChangeChannelPosImage(Image):
    axis_order = "TZCYX"


class TestInheritanceImageChannelPos(TestImageBase):
    image_class = ChangeChannelPosImage


class NoChannelImage(Image):
    axis_order = "TZYX"


class TestInheritanceNoChannelImage(TestImageBase):
    image_class = NoChannelImage

    def prepare_image_initial_shape(self, shape, channel):
        return self.prepare_mask_shape(shape)

    def needed_layer_shape(self, shape, axes: str, drop: str):
        axes = axes.replace("C", "")
        return super().needed_layer_shape(shape, axes, drop)


class ChangeTimePosImage(Image):
    axis_order = "ZTYXC"


class TestInheritanceImageTimePos(TestImageBase):
    image_class = ChangeTimePosImage


class WeirdOrderImage(Image):
    axis_order = "XYZTC"


class TestInheritanceWeirdOrderImage(TestImageBase):
    image_class = WeirdOrderImage


class AdditionalAxesImage(Image):
    axis_order = "VTZYXC"

    def get_image_for_save(self, index=0) -> np.ndarray:
        return super().get_image_for_save()[..., index]


class TestInheritanceAdditionalAxesImage(TestImageBase):
    image_class = AdditionalAxesImage


class TestMergeImage:
    @pytest.mark.parametrize("check_dtype", [np.uint8, np.uint16, np.uint32, np.float16, np.float32, np.float64])
    def test_merge_chanel(self, check_dtype):
        image1 = Image(data=np.zeros((3, 10, 10), dtype=np.uint8), axes_order="ZXY", spacing=(1, 1, 1))
        image2 = Image(data=np.ones((3, 10, 10), dtype=check_dtype), axes_order="ZXY", spacing=(1, 1, 1))
        res_image = image1.merge(image2, "C")
        assert res_image.channels == 2
        assert np.all(res_image.get_channel(0) == 0)
        assert np.all(res_image.get_channel(1) == 1)
        assert np.all(res_image.get_channel(Channel(1)) == 1)
        assert res_image.dtype == check_dtype

    def test_merge_fail(self):
        image1 = Image(data=np.zeros((4, 10, 10), dtype=np.uint8), axes_order="ZXY", spacing=(1, 1, 1))
        image2 = Image(data=np.zeros((3, 10, 10), dtype=np.uint8), axes_order="ZXY", spacing=(1, 1, 1))
        with pytest.raises(ValueError, match="Shape of arrays are different"):
            image1.merge(image2, "C")

    @pytest.mark.parametrize("axis_mark", Image.axis_order)
    def test_merge_different_axes(self, axis_mark):
        base_shape = (1, 1, 3, 10, 10)
        image1 = Image(data=np.zeros(base_shape, dtype=np.uint8), axes_order="CTZXY", spacing=(1, 1, 1))
        image2 = Image(data=np.ones(base_shape, dtype=np.uint8), axes_order="CTZXY", spacing=(1, 1, 1))
        res_image = image1.merge(image2, axis_mark)
        res_data = res_image.get_data()
        new_shape_li = list(base_shape)
        new_shape_li[Image.axis_order.index(axis_mark)] = new_shape_li[Image.axis_order.index(axis_mark)] * 2
        assert res_data.shape == tuple(new_shape_li)

    def test_merge_channel_name(self):
        image1 = Image(
            data=np.zeros((2, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel 1"), ChannelInfo(name="channel 3")],
        )
        image2 = Image(
            data=np.zeros((4, 10, 10), dtype=np.uint8),
            axes_order="ZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel 1")],
        )
        image3 = Image(
            data=np.zeros((4, 10, 10), dtype=np.uint8),
            axes_order="ZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel 5")],
        )
        res_image = image1.merge(image2, "C")
        assert res_image.channel_names == ["channel 1", "channel 3", "channel 1 (1)"]
        res_image = image1.merge(image3, "C")
        assert res_image.channel_names == ["channel 1", "channel 3", "channel 5"]

    def test_channel_name_form_str(self):
        image1 = Image(
            data=np.zeros((1, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel 1")],
        )
        assert image1.channel_names == ["channel 1"]
        image2 = Image(
            data=np.zeros((2, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel")],
        )
        assert image2.channel_names == ["channel", "channel 2"]
        image3 = Image(
            data=np.zeros((2, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_info=[ChannelInfo(name="channel 1")],
        )
        assert image3.channel_names == ["channel 1", "channel 2"]

    def test_different_axes_order(self):
        image1 = Image(data=np.zeros((3, 10, 10), dtype=np.uint8), axes_order="ZXY", spacing=(1, 1, 1))
        image2 = ChangeChannelPosImage(data=np.zeros((3, 10, 10), dtype=np.uint8), axes_order="ZXY", spacing=(1, 1, 1))
        res_image = image1.merge(image2, "C")
        assert res_image.channels == 2
        assert isinstance(res_image, Image)
        assert isinstance(image2.merge(image1, "C"), ChangeChannelPosImage)

    def test_get_data_by_axis(self):
        image = Image(data=np.zeros((2, 3, 10, 10), dtype=np.uint8), axes_order="CZXY", spacing=(1, 1, 1))
        assert image.get_data_by_axis(c=0).shape == (1, 3, 10, 10)
        assert image.get_data_by_axis(C=1).shape == (1, 3, 10, 10)
        assert image.get_data_by_axis(z=0).shape == (2, 1, 10, 10)
        assert image.get_data_by_axis(Z=1).shape == (2, 1, 10, 10)
        assert image.get_data_by_axis(C="channel 1").shape == (1, 3, 10, 10)


def test_cut_with_roi():
    data = np.zeros((3, 23, 23), np.uint8)
    data[0] = 1
    data[1] = 2
    data[2] = 3
    diam = diamond(5, dtype=np.uint8)
    mask = np.zeros((23, 23), np.uint8)
    mask[0:11, 0:11][diam > 0] = 1
    mask[6:17, 6:17][diam > 0] = 2
    mask[12:23, 12:23][diam > 0] = 3
    image = Image(data, spacing=(1, 1), axes_order="CXY")
    for i in range(1, 4):
        cut_image, cut_mask = image._cut_with_roi(mask == i, replace_mask=True, frame=2)
        assert cut_image[0].shape == (1, 1, 15, 15)
        assert np.all(cut_image[0][0, 0, 2:-2, 2:-2][diam > 0])
        assert np.all(cut_mask[0, 0, 2:-2, 2:-2] == diam)


def test_str_and_repr_mask_presence():
    image = Image(np.zeros((10, 10), np.uint8), spacing=(1, 1), file_path="test", axes_order="XY")
    assert "mask: False" in str(image)
    assert "mask=False" in repr(image)
    assert "coloring: ['red']" in str(image)

    image.set_mask(np.zeros((10, 10), np.uint8))

    assert "mask: True" in str(image)
    assert "mask=True" in repr(image)


def test_image_to_spacingspacing_rename():
    with pytest.warns(match="Argument image_spacing is deprecated since 0.15.4. Use spacing instead"):
        img = Image(np.zeros((10, 10), np.uint8), image_spacing=(2, 3), file_path="test", axes_order="XY")
    assert img.spacing == (2, 3)


def test_image_positional_to_named():
    with pytest.warns(match="Since PartSeg 0.15.4 all arguments, except first one, should be named"):
        img = Image(np.zeros((10, 10), np.uint8), (2, 3), "test", axes_order="XY")
    assert img.spacing == (2, 3)
    assert img.file_path == "test"


def test_merge_channel_props():
    with pytest.warns(match="Using channel_names, default_coloring and ranges is deprecated since PartSeg 0.15.4"):
        img = Image(
            data=np.zeros((2, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_names=["channel 2", "strange"],
            default_coloring=["red", (128, 255, 0)],
            ranges=[(0, 255), (0, 128)],
        )

    assert img.channel_names == ["channel 2", "strange"]
    assert img.default_coloring[0] == "red"
    assert tuple(img.default_coloring[1]) == (128, 255, 0)
    assert img.ranges == [(0, 255), (0, 128)]


@pytest.mark.parametrize(
    ("channel_name", "default_coloring", "ranges"),
    [
        ("channel", None, None),
        (
            None,
            [
                "blue",
            ],
            None,
        ),
        (None, None, [(0, 128)]),
        (["channel"], [], []),
    ],
)
def test_merge_channel_props_with_none(channel_name, default_coloring, ranges):
    with pytest.warns(match="Using channel_names, default_coloring and ranges is deprecated since PartSeg 0.15.4"):
        img = Image(
            data=np.zeros((1, 4, 10, 10), dtype=np.uint8),
            axes_order="CZXY",
            spacing=(1, 1, 1),
            channel_names=channel_name,
            default_coloring=default_coloring,
            ranges=ranges,
        )
    assert img.default_coloring == (default_coloring or ["red"])
    assert img.ranges == (ranges or [(0, 0)])
    assert img.channel_names == (["channel"] if channel_name else ["channel 1"])


def test_hex_to_rgb():
    assert _hex_to_rgb("#ff0000") == (255, 0, 0)
    assert _hex_to_rgb("#ff0000ff") == (255, 0, 0)
    assert _hex_to_rgb("#00FF00") == (0, 255, 0)
    assert _hex_to_rgb("#b00") == (187, 0, 0)
    assert _hex_to_rgb("#b00f") == (187, 0, 0)
    assert _hex_to_rgb("#B00") == (187, 0, 0)
    with pytest.raises(ValueError, match="Invalid hex code format"):
        _hex_to_rgb("#b0000")


def test_name_to_rgb():
    assert _name_to_rgb("red") == (255, 0, 0)
    assert _name_to_rgb("Red") == (255, 0, 0)
    assert _name_to_rgb("RED") == (255, 0, 0)
    assert _name_to_rgb("blue") == (0, 0, 255)
    assert _name_to_rgb("green") == (0, 128, 0)
    assert _name_to_rgb("white") == (255, 255, 255)
    assert _name_to_rgb("black") == (0, 0, 0)
    assert _name_to_rgb("yellow") == (255, 255, 0)
    with pytest.raises(ValueError, match="Unknown color name"):
        _name_to_rgb("strange")
    with pytest.raises(ValueError, match="Unknown color name"):
        _name_to_rgb("")


def test_name_to_rgb_vispy():
    # This test check mapping not defined in fallback dictionary
    pytest.importorskip("vispy", reason="vispy not installed")
    assert _name_to_rgb("lime") == (0, 255, 0)



from collections.abc import Sequence
from importlib.metadata import version

from napari import Viewer
from napari.layers import Labels
from packaging.version import parse as parse_version
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QTabWidget

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.label_create import LabelChoose, LabelEditor, LabelShow
from PartSeg.plugins.napari_widgets._settings import get_settings

NAPARI_GE_5_0 = parse_version(version("napari")) >= parse_version("0.5.0a1")
NAPARI_GE_4_19 = parse_version(version("napari")) >= parse_version("0.4.19a1")


class NapariLabelShow(LabelShow):
    def __init__(self, viewer: Viewer, name: str, label: list[Sequence[float]], removable, parent=None):
        super().__init__(name, label, removable, parent)
        self.viewer = viewer

        self.apply_label_btn = QPushButton("Apply")

        layout: QHBoxLayout = self.layout()
        layout.removeWidget(self.radio_btn)
        layout.insertWidget(0, self.apply_label_btn)
        viewer.layers.selection.events.changed.connect(self.update_preview)
        self.apply_label_btn.clicked.connect(self.apply_label)

    def update_preview(self, _event=None):
        if len(self.viewer.layers.selection) == 1 and isinstance(next(iter(self.viewer.layers.selection)), Labels):
            self.apply_label_btn.setEnabled(True)
            self.apply_label_btn.setToolTip("Apply labels to selected layer")
        else:
            self.apply_label_btn.setEnabled(False)
            self.apply_label_btn.setToolTip("Select one labels layer to apply custom labels")

    def apply_label(self):
        if len(self.viewer.layers.selection) == 1 and isinstance(
            layer := next(iter(self.viewer.layers.selection)), Labels
        ):
            max_val = layer.data.max()
            labels = {i + 1: [x / 255 for x in self.label[i % len(self.label)]] for i in range(max_val + 5)}
            labels[None] = [0, 0, 0, 0]
            if NAPARI_GE_4_19:
                from napari.utils.colormaps import DirectLabelColormap

                layer.colormap = DirectLabelColormap(color_dict=labels)
            else:
                layer.color = labels


class NapariLabelChoose(LabelChoose):
    def __init__(self, viewer: Viewer, settings, parent=None):
        super().__init__(settings, parent)
        self.viewer = viewer

    def _label_show(self, name: str, label: list[Sequence[float]], removable) -> LabelShow:
        return NapariLabelShow(self.viewer, name, label, removable, self)


class NapariLabelEditor(LabelEditor):
    settings: BaseSettings

    def save(self):
        super().save()
        self.settings.dump()


class LabelSelector(QTabWidget):
    def __init__(self, viewer: Viewer, parent=None):
        super().__init__(parent)
        self.viewer = viewer
        settings = get_settings()
        self.settings = settings
        self.label_editor = NapariLabelEditor(settings)
        self.label_view = NapariLabelChoose(viewer, settings)
        self.addTab(self.label_view, "Select labels")
        self.addTab(self.label_editor, "Create labels")

        self.label_view.edit_with_name_signal.connect(self.label_editor.set_colors)
        self.label_view.edit_signal.connect(self._set_label_editor)

    def _set_label_editor(self):
        self.setCurrentWidget(self.label_editor)



import warnings
from typing import Optional

import numpy as np
from magicgui.widgets import create_widget
from napari import Viewer
from napari.layers import Labels
from qtpy.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout, QWidget

from PartSeg.common_gui.mask_widget import MaskWidget
from PartSeg.plugins.napari_widgets._settings import get_settings
from PartSegCore.mask_create import calculate_mask
from PartSegCore.segmentation.algorithm_base import calculate_operation_radius


class NapariMaskWidget(MaskWidget):
    def __init__(self, settings, label_widget):
        super().__init__(settings)
        self.label_widget = label_widget

    def get_dilate_radius(self):
        if self.label_widget.value is None:
            return self.dilate_radius.value()
        spacing = self.label_widget.value.scale[-3:]
        radius = calculate_operation_radius(self.dilate_radius.value(), spacing, self.dilate_dim.currentEnum())
        if isinstance(radius, (list, tuple)):
            return [int(x + 0.5) for x in radius]
        return int(radius)


class MaskCreate(QWidget):
    def __init__(self, napari_viewer: Viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.settings = get_settings()
        self.roi_select = create_widget(annotation=Labels, label="ROI", options={})
        self.mask_select = create_widget(annotation=Optional[Labels], label="Base mask", options={})
        self.mask_widget = NapariMaskWidget(self.settings, self.roi_select)
        self.create = QPushButton("Create")
        self.name = QLineEdit()
        self.name.setText(self.settings.get("mask_create_name", "Mask"))
        layout = QVBoxLayout()
        layout2 = QHBoxLayout()
        layout2.addWidget(QLabel("ROI"))
        layout2.addWidget(self.roi_select.native)
        layout.addLayout(layout2)
        layout2 = QHBoxLayout()
        layout2.addWidget(QLabel("Base mask"))
        layout2.addWidget(self.mask_select.native)
        layout.addLayout(layout2)
        layout2 = QHBoxLayout()
        layout2.addWidget(QLabel("New layer name:"))
        layout2.addWidget(self.name)
        layout.addLayout(layout2)
        layout.addWidget(self.mask_widget)
        layout.addWidget(self.mask_widget.radius_information)
        layout.addWidget(self.create)
        self.setLayout(layout)

        self.mask_select.native.setDisabled(True)
        self.mask_widget.clip_to_mask.stateChanged.connect(self.mask_select.native.setEnabled)
        self.create.clicked.connect(self.create_mask)

    def showEvent(self, event) -> None:
        super().showEvent(event)
        self.reset_choices()

    def reset_choices(self, event=None):
        self.roi_select.reset_choices(event)
        self.mask_select.reset_choices(event)

    def create_mask(self):
        if self.roi_select.value is None:
            return
        layer_name = self.name.text()
        self.settings.set("mask_create_name", layer_name)
        mask_property = self.mask_widget.get_mask_property()
        if mask_property.clip_to_mask and self.mask_select.value is None:
            warnings.warn("Select base mask", RuntimeWarning, stacklevel=1)
            return
        base_mask = None if self.mask_select.value is None else self.mask_select.value.data
        scale = np.array(self.roi_select.value.scale)
        mask = np.array(calculate_mask(mask_property, self.roi_select.value.data, base_mask, scale[-3:]))
        if layer_name in self.viewer.layers:
            self.viewer.layers[layer_name].data = mask
        else:
            self.viewer.add_labels(
                mask,
                scale=scale[-int(mask.ndim) :],
                name=layer_name,
            )

        self.settings.dump()



from __future__ import annotations

import re
import sys
import typing
import warnings
from contextlib import suppress
from copy import copy
from dataclasses import dataclass
from functools import wraps
from itertools import cycle, zip_longest

import numpy as np

from PartSegImage.channel_class import Channel

Spacing = tuple[typing.Union[float, int], ...]
_IMAGE_DATA = typing.Union[list[np.ndarray], np.ndarray]

_DEF = object()
FRAME_THICKNESS = 2

DEFAULT_SCALE_FACTOR = 10**9

ch_par: dict[str, bool]

if sys.version_info[:2] > (3, 9):
    ch_par = {"kw_only": True, "slots": True}
else:
    ch_par = {}


@dataclass(**ch_par)
class ChannelInfo:
    name: str
    color_map: str | np.ndarray | tuple | list | None = None
    contrast_limits: tuple[float, float] | None = None


@dataclass(**ch_par)
class ChannelInfoFull:
    name: str
    color_map: str | np.ndarray
    contrast_limits: tuple[float, float]

    def __post_init__(self):
        if not isinstance(self.color_map, (str, np.ndarray)):
            self.color_map = np.array(self.color_map)


def minimal_dtype(val: int):
    """
    Calculate minimal type to handle value in array

    :param val:
    :return: minimal dtype to handle given value
    :rtype:
    """
    if val < 250:
        return np.uint8
    return np.uint16 if val < 2**16 - 5 else np.uint32


def reduce_array(
    array: np.ndarray,
    components: typing.Collection[int] | None = None,
    max_val: int | None = None,
    dtype=None,
) -> np.ndarray:
    """
    Relabel components from 1 to components_num with keeping order

    :param array: array to relabel, deed to be integer type
    :param components: components to be keep, if None then all will be keep
    :param max_val: number of maximum component in array, if absent then will be calculated
        (to reduce whole array processing)
    :param dtype: type of returned array if no then minimal type is calculated
    :return: relabeled array in minimum type
    """
    # this function minimal dtype is np.uint8 so there is no need to do calculation.
    if components is None:
        components = np.unique(array.flat)
        if max_val is None:
            max_val = np.max(components)

    if max_val is None:
        max_val = np.max(array)

    translate = np.zeros(max_val + 1, dtype=dtype or minimal_dtype(len(components) + 1))

    for i, val in enumerate(sorted(components), start=0 if 0 in components else 1):
        translate[val] = i

    return translate[array]


def rename_argument(from_name: str, to_name: str, since_version: str):
    def decorator(fun):
        @wraps(fun)
        def _fun(*args, **kwargs):
            if from_name in kwargs:
                warnings.warn(
                    f"Argument {from_name} is deprecated since {since_version}. Use {to_name} instead",
                    DeprecationWarning,
                    stacklevel=2,
                )
                kwargs[to_name] = kwargs.pop(from_name)
            return fun(*args, **kwargs)

        return _fun

    return decorator


def positional_to_named(fun):
    @wraps(fun)
    def _fun(*args, **kwargs):
        if len(args) > 2:
            warnings.warn(
                "Since PartSeg 0.15.4 all arguments, except first one, should be named",
                DeprecationWarning,
                stacklevel=2,
            )

        new_kwargs = dict(
            zip(
                (
                    "spacing",
                    "file_path",
                    "mask",
                    "default_coloring",
                    "ranges",
                    "axes_order",
                    "shift",
                    "name",
                    "metadata_dict",
                ),
                args[2:],
                # start from 2 because first two arguments are self and data
            )
        )

        kwargs.update(new_kwargs)

        return fun(*args[:2], **kwargs)

    return _fun


def merge_into_channel_info(fun):
    @wraps(fun)
    def _fun(*args, **kwargs):
        if "channel_info" in kwargs:
            fun(*args, **kwargs)
            return None
        channel_names = kwargs.pop("channel_names", [])
        default_coloring = kwargs.pop("default_coloring", [])
        ranges = kwargs.pop("ranges", [])
        if any([channel_names, default_coloring, ranges]):
            if isinstance(channel_names, str):
                channel_names = [channel_names]
            if channel_names is None:
                channel_names = []
            if default_coloring is None:
                default_coloring = []
            if ranges is None:
                ranges = []
            warnings.warn(
                "Using channel_names, default_coloring and ranges is deprecated since PartSeg 0.15.4",
                DeprecationWarning,
                stacklevel=2,
            )
            channel_info = [
                ChannelInfo(name=name, color_map=color, contrast_limits=contrast_limits)
                for name, color, contrast_limits in zip_longest(channel_names, default_coloring, ranges)
            ]
            kwargs["channel_info"] = channel_info
        return fun(*args, **kwargs)

    return _fun


class Image:
    """
    Base class for Images used in PartSeg

    :param data: 5-dim array with order: time, z, y, x, channel
    :param spacing: spacing for z, y, x
    :param file_path: path to image on disc
    :param mask: mask array in shape z,y,x
    :param channel_info: list of metadata stored per channel
    :param axes_order: allow to create Image object form data with different axes order, or missed axes

    :cvar str ~.axis_order: internal order of axes

    It is prepared for subclassing with changed internal order. Eg:

    >>> class ImageJImage(Image):
    >>>     axis_order = "TZCYX"

    """

    _image_spacing: Spacing
    axis_order = "CTZYX"
    array_axis_order: str

    def __new__(cls, *args, **kwargs):
        if hasattr(cls, "return_order"):  # pragma: no cover
            warnings.warn("Using return_order is deprecated since PartSeg 0.11.0", DeprecationWarning, stacklevel=2)
            cls.axis_order = cls.return_order
        cls.array_axis_order = cls.axis_order.replace("C", "")
        return super().__new__(cls)

    @positional_to_named
    @rename_argument("image_spacing", "spacing", "0.15.4")
    @merge_into_channel_info
    def __init__(
        self,
        data: _IMAGE_DATA,
        *,
        spacing: Spacing,
        file_path=None,
        mask: None | np.ndarray = None,
        channel_info: list[ChannelInfo | ChannelInfoFull] | None = None,
        axes_order: str | None = None,
        shift: Spacing | None = None,
        name: str = "",
        metadata_dict: dict | None = None,
    ):
        # TODO add time distance to image spacing
        if axes_order is None:  # pragma: no cover
            warnings.warn(
                f"axes_order should be provided, Currently it uses {self.__class__}.axis_order",
                category=DeprecationWarning,
                stacklevel=2,
            )
            axes_order = self.axis_order
        self._check_data_dimensionality(data, axes_order)
        if not isinstance(spacing, tuple):
            spacing = tuple(spacing)
        self._channel_arrays = self._split_data_on_channels(data, axes_order)
        self._image_spacing = (1.0,) * (3 - len(spacing)) + spacing
        self._image_spacing = tuple(el if el > 0 else 10**-6 for el in self._image_spacing)

        self._shift = tuple(shift) if shift is not None else (0,) * len(self._image_spacing)
        self.name = name

        self.file_path = file_path

        self._mask_array = self._fit_mask(mask, data, axes_order)
        self._channel_info = self._adjust_channel_info(channel_info, self._channel_arrays)
        self.metadata = dict(metadata_dict) if metadata_dict is not None else {}

    @staticmethod
    def _adjust_channel_info(
        channel_info: list[ChannelInfo | ChannelInfoFull] | None,
        channel_array: list[np.ndarray],
        default_colors=("red", "blue", "green", "yellow", "magenta", "cyan"),
    ) -> list[ChannelInfoFull]:
        default_colors = cycle(default_colors)
        if channel_info is None:
            ranges = [(np.min(x), np.max(x)) for x in channel_array]
            return [
                ChannelInfoFull(name=f"channel {i}", color_map=x[0], contrast_limits=x[1])
                for i, x in enumerate(zip(default_colors, ranges), start=1)
            ]

        channel_info = channel_info[: len(channel_array)]

        res = [
            ChannelInfoFull(
                name=ch_inf.name or f"channel {i+1}",
                color_map=(
                    ch_inf.color_map if ch_inf.color_map is not None else next(default_colors)  # skipcq: PTC-W0063
                ),
                contrast_limits=(
                    ch_inf.contrast_limits
                    if ch_inf.contrast_limits is not None
                    else (np.min(channel_array[i]), np.max(channel_array[i]))
                ),
            )
            for i, ch_inf in enumerate(channel_info)
        ]
        res.extend(
            ChannelInfoFull(
                name=f"channel {i+1}",
                color_map=next(default_colors),  # skipcq: PTC-W0063
                contrast_limits=(np.min(arr), np.max(arr)),
            )
            for i, arr in enumerate(channel_array[len(res) :], start=len(channel_info))
        )
        return res

    @staticmethod
    def _check_data_dimensionality(data, axes_order):
        if (isinstance(data, list) and any(x.ndim + 1 != len(axes_order) for x in data)) or (
            not isinstance(data, list) and data.ndim != len(axes_order)
        ):
            if isinstance(data, list):
                ndim = ", ".join([f"{x.ndim} + 1" for x in data])
            else:
                ndim = str(data.ndim)
            raise ValueError(
                "Data should have same number of dimensions "
                f"like length of axes_order (axis :{len(axes_order)}, ndim: {ndim}"
            )

    def _fit_mask(self, mask, data, axes_order):
        mask_array = self._prepare_mask(mask, data, axes_order)
        if mask_array is not None:
            mask_array = self.fit_mask_to_image(mask_array)
        return mask_array

    @classmethod
    def _prepare_mask(cls, mask, data, axes_order) -> np.ndarray | None:
        if mask is None:
            return None

        if isinstance(data, list):
            data_shape = list(data[0].shape)
        else:
            data_shape = list(data.shape)
            with suppress(ValueError):
                data_shape.pop(axes_order.index("C"))

        mask = cls._fit_array_to_image(data_shape, mask)
        return cls.reorder_axes(mask, axes_order.replace("C", ""))

    @classmethod
    def _split_data_on_channels(cls, data: np.ndarray | list[np.ndarray], axes_order: str) -> list[np.ndarray]:
        if isinstance(data, list) and not axes_order.startswith("C"):  # pragma: no cover
            raise ValueError("When passing data as list of numpy arrays then Channel must be first axis.")
        if "C" not in axes_order:
            if not isinstance(data, np.ndarray):  # pragma: no cover
                raise TypeError("If `axes_order` does not contain `C` then data must be numpy array.")
            return [cls.reorder_axes(data, axes_order)]
        if axes_order.startswith("C"):
            if isinstance(data, list):
                dtype = np.result_type(*data)
                return [cls.reorder_axes(x, axes_order[1:]).astype(dtype) for x in data]
            return [cls.reorder_axes(x, axes_order[1:]) for x in data]

        if not isinstance(data, np.ndarray):
            raise TypeError("If `data` is list of arrays then `axes_order` must start with `C`")  # pragma: no cover
        pos: list[slice | int] = [slice(None) for _ in range(data.ndim)]
        c_pos = axes_order.index("C")
        res = []
        for i in range(data.shape[c_pos]):
            pos[c_pos] = i
            res.append(cls.reorder_axes(data[tuple(pos)], axes_order.replace("C", "")))
        return res

    @staticmethod
    def _merge_channel_names(base_channel_names: list[str], new_channel_names: list[str]) -> list[str]:
        base_channel_names = base_channel_names[:]
        reg = re.compile(r"channel \d+")
        for name in new_channel_names:
            match = reg.match(name)
            new_name = name
            base_name = name
            if match and base_name in base_channel_names:
                new_name = f"channel {len(base_channel_names) + 1}"
            i = 1
            while new_name in base_channel_names:
                new_name = f"{base_name} ({i})"
                i += 1
                if i > 10000:  # pragma: no cover
                    raise ValueError("fail when try to fix channel name")
            base_channel_names.append(new_name)
        return base_channel_names

    @property
    def channel_info(self) -> list[ChannelInfoFull]:
        return [copy(x) for x in self._channel_info]

    @property
    def ranges(self) -> list[tuple[float, float]]:
        return [x.contrast_limits for x in self._channel_info]

    @property
    def default_coloring(self) -> list[str | np.ndarray]:
        return [x.color_map for x in self._channel_info]

    def merge(self, image: Image, axis: str) -> Image:
        """
        Produce new image merging image data along given axis. All metadata
        are obtained from self.

        :param Image image: Image to be merged
        :param str axis:
        :return: New image produced from merge
        :rtype: Image
        """
        if axis == "C":
            data = self._image_data_normalize(
                self._channel_arrays + [self.reorder_axes(x, image.array_axis_order) for x in image._channel_arrays]
            )
            channel_names = self._merge_channel_names(self.channel_names, image.channel_names)
            color_map = self.default_coloring + image.default_coloring
        else:
            index = self.array_axis_order.index(axis)
            data = self._image_data_normalize(
                [
                    np.concatenate((y, self.reorder_axes(y, image.array_axis_order)), axis=index)
                    for x, y in zip(self._channel_arrays, image._channel_arrays)
                ]
            )
            channel_names = self.channel_names
            color_map = self.default_coloring

        return self.substitute(
            data=data, ranges=self.ranges + image.ranges, channel_names=channel_names, default_coloring=color_map
        )

    @property
    def channel_names(self) -> list[str]:
        return [x.name for x in self._channel_info]

    @property
    def channel_pos(self) -> int:  # pragma: no cover
        """Channel axis. Need to have 'C' in :py:attr:`axis_order`"""
        warnings.warn(
            "channel_pos is deprecated and code its using may not work properly", category=FutureWarning, stacklevel=2
        )
        return self.axis_order.index("C")

    @property
    def x_pos(self):
        return self.array_axis_order.index("X")

    @property
    def y_pos(self):
        return self.array_axis_order.index("Y")

    @property
    def time_pos(self):
        """Time axis. Need to have 'T' in :py:attr:`axis_order`"""
        return self.array_axis_order.index("T")

    @property
    def stack_pos(self) -> int:
        """Stack axis. Need to have 'Z' in :py:attr:`axis_order`"""
        return self.array_axis_order.index("Z")

    @property
    def dtype(self) -> np.dtype:
        """dtype of image array"""
        return self._channel_arrays[0].dtype

    @staticmethod
    def _reorder_axes(array: np.ndarray, input_axes: str, return_axes) -> np.ndarray:
        if array.ndim != len(input_axes):
            raise ValueError(f"array.ndim ({array.ndim}) need to be equal to length of axes ('{input_axes}')")
        if input_axes == return_axes:
            return array
        mapping_dict = {v: i for i, v in enumerate(return_axes)}
        if array.ndim < len(return_axes):
            array = array.reshape(array.shape + (1,) * (len(return_axes) - array.ndim))
        new_positions = [mapping_dict[x] for x in input_axes if x in mapping_dict]
        axes_to_map = [i for i, x in enumerate(input_axes) if x in mapping_dict]
        return np.moveaxis(array, axes_to_map, new_positions)

    @classmethod
    def reorder_axes(cls, array: np.ndarray, axes: str) -> np.ndarray:
        """
        reorder axes to internal storage format

        :param np.ndarray array: array to have changed order of axes
        :param str axes: axes order
        :return: array with correct order of axes
        """
        return cls._reorder_axes(array, axes, cls.array_axis_order)

    def get_dimension_number(self) -> int:
        """return number of nontrivial dimensions"""
        return np.squeeze(self._channel_arrays[0]).ndim

    def get_dimension_letters(self) -> str:
        """
        :return: letters which indicates non trivial dimensions
        """
        return "".join(key for val, key in zip(self._channel_arrays[0].shape, self.array_axis_order) if val > 1)

    def substitute(
        self,
        data=None,
        image_spacing=None,
        file_path=None,
        mask=_DEF,
        default_coloring=None,
        ranges=None,
        channel_names=None,
    ) -> Image:
        """Create copy of image with substitution of not None elements"""
        data = self._channel_arrays if data is None else data
        image_spacing = self._image_spacing if image_spacing is None else image_spacing
        file_path = self.file_path if file_path is None else file_path
        mask = self._mask_array if mask is _DEF else mask
        default_coloring = self.default_coloring if default_coloring is None else default_coloring
        ranges = self.ranges if ranges is None else ranges
        channel_names = self.channel_names if channel_names is None else channel_names

        channel_info = [
            ChannelInfo(name=name, color_map=color, contrast_limits=contrast_limits)
            for name, color, contrast_limits in zip_longest(channel_names, default_coloring, ranges)
        ]

        return self.__class__(
            data=data,
            spacing=image_spacing,
            file_path=file_path,
            mask=mask,
            axes_order=self.axis_order,
            channel_info=channel_info,
            metadata_dict=self.metadata,
        )

    def set_mask(self, mask: np.ndarray | None, axes: str | None = None):
        """
        Set mask for image, check if it has proper shape.

        :param mask: mask in same shape like image. May not contains 1 dim axes.
        :param axes: order of axes in mask, use if different than :py:attr:`return_order`
        :raise ValueError: on wrong shape
        """
        if mask is None:
            self._mask_array = None
        elif axes is not None:
            self._mask_array = self.fit_mask_to_image(self.reorder_axes(mask, axes))
        else:
            self._mask_array = self.fit_mask_to_image(mask)

    def get_data(self) -> np.ndarray:
        if "C" in self.axis_order:
            return np.stack(self._channel_arrays, axis=self.axis_order.index("C"))
        return self._channel_arrays[0]

    @property
    def mask(self) -> np.ndarray | None:
        return self._mask_array[:] if self._mask_array is not None else None

    @staticmethod
    def _fit_array_to_image(base_shape, array: np.ndarray) -> np.ndarray:
        """change shape of array with inserting single dimensional entries"""
        shape = list(array.shape)
        for i, el in enumerate(base_shape):
            if el == 1 and el != shape[i]:
                shape.insert(i, 1)
            elif el != shape[i]:
                raise ValueError(f"Wrong array shape {shape} for {base_shape}")
        if len(shape) != len(base_shape):
            raise ValueError(f"Wrong array shape {shape} for {base_shape}")
        return np.reshape(array, shape)

    def fit_array_to_image(self, array: np.ndarray) -> np.ndarray:
        """
        Change shape of array with inserting single dimensional entries

        :param np.ndarray array: array to be fitted

        :return: reshaped array with added missing 1 in shape

        :raises ValueError: if cannot fit array
        """
        return self._fit_array_to_image(self._channel_arrays[0].shape, array)

    # noinspection DuplicatedCode
    def fit_mask_to_image(self, array: np.ndarray) -> np.ndarray:
        """
        call :py:meth:`fit_array_to_image` and
        then relabel and change type to minimal
        which fit all information
        """
        array = self.fit_array_to_image(array)
        if np.max(array) == 1:
            return array.astype(np.uint8)
        unique = np.unique(array)
        if unique.size == 1:
            if unique[0] != 0:
                return np.ones(array.shape, dtype=np.uint8)
            return array.astype(np.uint8)
        max_val = np.max(unique)
        return reduce_array(array, unique, max_val)

    def get_image_for_save(self) -> np.ndarray:
        """
        :return: numpy array in imagej tiff order axes
        """
        if "C" in self.axis_order:
            return self._reorder_axes(
                np.stack(self._channel_arrays, axis=self.axis_order.index("C")), self.axis_order, "TZCYX"
            )
        return self._reorder_axes(self._channel_arrays[0], self.axis_order, "TZCYX")

    def get_mask_for_save(self) -> np.ndarray | None:
        """
        :return: if image has mask then return mask with axes in proper order
        """
        if self._mask_array is None:
            return None
        return self._reorder_axes(self._mask_array, "".join(self.array_axis_order), "TZCYX")

    @property
    def has_mask(self) -> bool:
        """check if image is masked"""
        return self._mask_array is not None

    @property
    def is_time(self) -> bool:
        """check if image contains time data"""
        return self.times > 1

    @property
    def is_stack(self) -> bool:
        """check if image contain 3d data"""
        return self.layers > 1

    @property
    def channels(self) -> int:
        """number of image channels"""
        return len(self._channel_arrays)

    @property
    def layers(self) -> int:
        """z-dim of image"""
        return self._channel_arrays[0].shape[self.stack_pos]

    @property
    def times(self) -> int:
        """number of time frames"""
        return self._channel_arrays[0].shape[self.time_pos]

    @property
    def plane_shape(self) -> tuple[int, int]:
        """y,x size of image"""
        return self._channel_arrays[0].shape[self.y_pos], self._channel_arrays[0].shape[self.x_pos]

    @property
    def shape(self):
        """Whole image shape. order of axes my change. Current order is in :py:attr:`return_order`"""
        return self._channel_arrays[0].shape

    def swap_time_and_stack(self):
        """
        Swap time and stack axes.
        For example my be used to convert time image in 3d image.
        """
        image_array_list = [np.swapaxes(x, self.time_pos, self.stack_pos) for x in self._channel_arrays]
        return self.substitute(data=self._image_data_normalize(image_array_list))

    @classmethod
    def get_axis_positions(cls) -> dict[str, int]:
        """
        :return: dict with mapping axis to its position
        :rtype: dict
        """
        return {letter: i for i, letter in enumerate(cls.axis_order)}

    @classmethod
    def get_array_axis_positions(cls) -> dict[str, int]:
        """
        :return: dict with mapping axis to its position for array fitted to image
        :rtype: dict
        """
        return {letter: i for i, letter in enumerate(cls.array_axis_order)}

    def get_data_by_axis(self, **kwargs) -> np.ndarray:
        """
        Get part of data extracted by sub axis. Axis is selected by single letter from :py:attr:`axis_order`

        :param kwargs: axis list with
        :return:
        :rtype:
        """
        slices: list[int | slice] = [slice(None) for _ in range(len(self.array_axis_order))]
        axis_pos = self.get_array_axis_positions()
        if "c" in kwargs:
            kwargs["C"] = kwargs.pop("c")
        if "C" in kwargs:
            if isinstance(kwargs["C"], Channel):
                kwargs["C"] = kwargs["C"].value
            if isinstance(kwargs["C"], str):
                kwargs["C"] = self.channel_names.index(kwargs["C"])

        channel = kwargs.pop("C", slice(None) if "C" in self.axis_order else 0)
        if isinstance(channel, Channel):
            channel = channel.value

        axis_order = self.axis_order
        for name, value in kwargs.items():
            if name.upper() in axis_pos:
                slices[axis_pos[name.upper()]] = value
                if isinstance(value, int):
                    axis_order = axis_order.replace(name.upper(), "")

        slices_t = tuple(slices)
        if isinstance(channel, int):
            return self._channel_arrays[channel][slices_t]
        return np.stack([x[slices_t] for x in self._channel_arrays[channel]], axis=axis_order.index("C"))

    def clip_array(self, array: np.ndarray, **kwargs: int | slice) -> np.ndarray:
        """
        Clip array by axis. Axis is selected by single letter from :py:attr:`axis_order`

        :param array: array to clip
        :param kwargs: mapping from axis to position or slice on this axis
        :return: clipped array
        """
        array = self.fit_array_to_image(array)
        slices: list[int | slice] = [slice(None) for _ in range(len(self.array_axis_order))]
        axis_pos = self.get_array_axis_positions()
        for name, value in kwargs.items():
            if (n := name.upper()) in axis_pos:
                slices[axis_pos[n]] = value
        return array[tuple(slices)]

    def get_channel(self, num: int | str | Channel) -> np.ndarray:
        """
        Alias for :py:func:`get_sub_data` with argument ``c=num``

        :param int | str | Channel num: channel num or name to be extracted
        :return: given channel array
        :rtype: numpy.ndarray
        """
        return self.get_data_by_axis(c=num)

    def has_channel(self, num: int | str | Channel) -> bool:
        if isinstance(num, Channel):
            num = num.value

        if isinstance(num, str):
            return num in self.channel_names
        return 0 <= num < self.channels

    def get_layer(self, time: int, stack: int) -> np.ndarray:
        """
        return single layer contains data for all channel

        :param time: time coordinate. For images with not time use 0.
        :param stack: "z coordinate. For time data use 0.
        :return:
        """
        warnings.warn(
            "Image.get_layer is deprecated. Use get_data_by_axis instead", category=DeprecationWarning, stacklevel=2
        )
        return self.get_data_by_axis(T=time, Z=stack)

    @property
    def is_2d(self) -> bool:
        """
        Check if image z and time dimension are equal to 1.
        Equivalent to:
        `image.layers == 1 and image.times == 1`
        """
        return self.layers == 1 and self.times == 1

    @property
    def spacing(self) -> Spacing:
        """image spacing"""
        return tuple(self._image_spacing[1:]) if self.is_2d else self._image_spacing

    def normalized_scaling(self, factor=DEFAULT_SCALE_FACTOR) -> Spacing:
        if self.is_2d:
            return (1, 1, *tuple(np.multiply(self.spacing, factor)))
        return (1, *tuple(np.multiply(self.spacing, factor)))

    @property
    def shift(self):
        return self._shift[1:] if self.is_2d else self._shift

    @property
    def voxel_size(self) -> Spacing:
        """alias for spacing"""
        return self.spacing

    def set_spacing(self, value: Spacing):
        """set image spacing"""
        if 0 in value:
            return
        if self.is_2d and len(value) + 1 == len(self._image_spacing):
            value = (1.0, *tuple(value))
        if len(value) != len(self._image_spacing):  # pragma: no cover
            raise ValueError("Correction of spacing fail.")
        self._image_spacing = tuple(value)

    @staticmethod
    def _frame_array(array: np.ndarray | None, index_to_add: list[int], frame=FRAME_THICKNESS):
        if array is None:  # pragma: no cover
            return array
        result_shape = list(array.shape)
        image_pos = [slice(None) for _ in range(array.ndim)]

        for index in index_to_add:
            result_shape[index] += frame * 2
            image_pos[index] = slice(frame, result_shape[index] - frame)

        data = np.zeros(shape=result_shape, dtype=array.dtype)
        data[tuple(image_pos)] = array
        return data

    @staticmethod
    def calc_index_to_frame(array_axis: str, important_axis: str) -> list[int]:
        """
        calculate in which axis frame should be added

        :param str array_axis: list of image axis
        :param str important_axis: list of framed axis
        :return: list of indices to add frame.
        """
        return [array_axis.index(letter) for letter in important_axis]

    def _frame_cut_area(self, cut_area: typing.Iterable[slice], frame: int):
        cut_area = list(cut_area)
        important_axis = "XY" if self.is_2d else "XYZ"
        for ind in self.calc_index_to_frame(self.array_axis_order, important_axis):
            sl = cut_area[ind]
            cut_area[ind] = slice(
                max(sl.start - frame, 0) if sl.start is not None else None,
                sl.stop + frame if sl.stop is not None else None,
                sl.step,
            )
        return cut_area

    def _cut_image_slices(
        self, cut_area: typing.Iterable[slice], frame: int
    ) -> tuple[list[np.ndarray], np.ndarray | None]:
        new_mask = None
        cut_area = self._frame_cut_area(cut_area, frame)
        new_image = [x[tuple(cut_area)] for x in self._channel_arrays]
        if self._mask_array is not None:
            new_mask = self._mask_array[tuple(cut_area)]
        return new_image, new_mask

    def _roi_to_slices(self, roi: np.ndarray) -> list[slice]:
        cut_area = self.fit_array_to_image(roi)
        points = np.nonzero(cut_area)
        lower_bound = np.min(points, axis=1)
        upper_bound = np.max(points, axis=1)
        return [slice(x, y + 1) for x, y in zip(lower_bound, upper_bound)]

    def _cut_with_roi(self, cut_area: np.ndarray, replace_mask: bool, frame: int):
        new_mask = None
        cut_area = self.fit_array_to_image(cut_area)
        new_cut = tuple(self._roi_to_slices(cut_area))
        catted_cut_area = cut_area[new_cut]
        new_image = [np.copy(x[new_cut]) for x in self._channel_arrays]
        for el in new_image:
            el[catted_cut_area == 0] = 0
        if replace_mask:
            new_mask = catted_cut_area
        elif self._mask_array is not None:
            new_mask = self._mask_array[new_cut]
            new_mask[catted_cut_area == 0] = 0
        important_axis = "XY" if self.is_2d else "XYZ"
        new_image = [
            self._frame_array(x, self.calc_index_to_frame(self.array_axis_order, important_axis), frame)
            for x in new_image
        ]

        new_mask = self._frame_array(new_mask, self.calc_index_to_frame(self.array_axis_order, important_axis), frame)
        return new_image, new_mask

    def cut_image(
        self,
        cut_area: np.ndarray | typing.Iterable[slice],
        replace_mask=False,
        frame: int = FRAME_THICKNESS,
        zero_out_cut_area: bool = True,
    ) -> Image:
        """
        Create new image base on mask or list of slices
        :param bool replace_mask: if cut area is represented by mask array,
        then in result image the mask is set base on cut_area if cur_area is np.ndarray
        :param typing.Union[np.ndarray, typing.Iterable[slice]] cut_area: area to cut. Defined with slices or mask
        :param int frame: additional frame around cut_area
        :param bool zero_out_cut_area:
        :return: Image
        """
        if isinstance(cut_area, np.ndarray):
            if zero_out_cut_area:
                new_image, new_mask = self._cut_with_roi(cut_area, replace_mask, frame)
            else:
                new_cut = self._roi_to_slices(cut_area)
                new_image, new_mask = self._cut_image_slices(new_cut, frame)
                if replace_mask:
                    new_mask = cut_area[tuple(self._frame_cut_area(new_cut, frame))]
        else:
            new_image, new_mask = self._cut_image_slices(cut_area, frame)

        return self.__class__(
            data=self._image_data_normalize(new_image),
            spacing=self._image_spacing,
            file_path=None,
            mask=new_mask,
            channel_info=self._channel_info,
            axes_order=self.axis_order,
        )

    def get_imagej_colors(self):
        res = []
        for color in self.default_coloring:

            if isinstance(color, str):
                if color.startswith("#"):
                    color_array = _hex_to_rgb(color)
                else:
                    color_array = _name_to_rgb(color)
                res.append(np.array([np.linspace(0, x, num=256) for x in color_array]).astype(np.uint8))
            elif color.ndim == 1:
                res.append(np.array([np.linspace(0, x, num=256) for x in color]).astype(np.uint8))
            else:
                if color.shape[1] != 256:
                    res.append(
                        np.array(
                            [
                                np.interp(np.linspace(0, 255, num=256), np.linspace(0, color.shape[1], num=256), x)
                                for x in color
                            ]
                        )
                    )
                res.append(color)
        return res

    def get_colors(self) -> list[str | list[int]]:
        res: list[str | list[int]] = []
        for color in self.default_coloring:
            if isinstance(color, str):
                res.append(color)
            elif color.ndim == 2:
                res.append(list(color[:, -1]))
            else:
                res.append(list(color))
        return res

    def get_um_spacing(self) -> Spacing:
        """image spacing in micrometers"""
        return tuple(float(x * 10**6) for x in self.spacing)

    def get_um_shift(self) -> Spacing:
        """image spacing in micrometers"""
        return tuple(float(x * 10**6) for x in self.shift)

    def get_ranges(self) -> list[tuple[float, float]]:
        """image brightness ranges for each channel"""
        return self.ranges[:]

    def __str__(self):
        return (
            f"{self.__class__} Shape {self._channel_arrays[0].shape}, dtype: {self._channel_arrays[0].dtype}, "
            f"labels: {self.channel_names}, coloring: {self.get_colors()} mask: {self.has_mask}"
        )

    def __repr__(self):
        mask_info = f"mask=True, mask_dtype={self._mask_array.dtype}" if self.mask is not None else "mask=False"
        return (
            f"Image(shape={self._channel_arrays[0].shape} dtype={self._channel_arrays[0].dtype}, spacing={self.spacing}"
            f", labels={self.channel_names}, channels={self.channels}, axes={self.axis_order!r}, {mask_info})"
        )

    @classmethod
    def _image_data_normalize(cls, data: _IMAGE_DATA) -> _IMAGE_DATA:
        if isinstance(data, np.ndarray):
            return data
        if cls.axis_order.startswith("C"):
            shape = data[0].shape
            if any(x.shape != shape for x in data):
                raise ValueError(f"Shape of arrays are different {[x.shape for x in data]}")
            return data
        if "C" not in cls.axis_order:
            return data[0]
        return np.stack(data, axis=cls.axis_order.index("C"))


def _hex_to_rgb(hex_code: str) -> tuple[int, int, int]:
    """
    Convert a hex color code to an RGB tuple.

    :param str hex_code: The hex color code, either short form (#RGB) or long form (#RRGGBB)
    :return: A tuple containing the RGB values (R, G, B)
    """
    hex_code = hex_code.lstrip("#")

    if len(hex_code) in {3, 4}:
        hex_code = "".join([c * 2 for c in hex_code])
    elif len(hex_code) not in {6, 8}:
        raise ValueError(f"Invalid hex code format: {hex_code}")

    return int(hex_code[:2], 16), int(hex_code[2:4], 16), int(hex_code[4:6], 16)


def _name_to_rgb(name: str) -> tuple[int, int, int]:
    """
    Convert a color name to an RGB tuple.

    :param str name: The color name
    :return: A tuple containing the RGB values (R, G, B)
    """
    name = name.lower()
    if name not in _NAMED_COLORS:
        raise ValueError(f"Unknown color name: {name}")
    return _hex_to_rgb(_NAMED_COLORS[name])


try:
    from vispy.color import get_color_dict
except ImportError:  # pragma: no cover
    _NAMED_COLORS = {
        "red": "#FF0000",
        "green": "#008000",
        "blue": "#0000FF",
        "yellow": "#FFFF00",
        "cyan": "#00FFFF",
        "magenta": "#FF00FF",
        "white": "#FFFFFF",
        "black": "#000000",
        "orange": "#FFA500",
    }
else:
    _NAMED_COLORS = get_color_dict()



"""
Module with default values of application state
"""

import os
import sys

import appdirs
import packaging.version

from PartSeg import APP_LAB, APP_NAME, __version__, parsed_version

#: if true then report errors using sentry without asking user
auto_report = parsed_version.is_devrelease and getattr(sys, "frozen", False)
#: if report errors using sentry without asking user about it
always_report = False
#: if enable error reporting button (to allow see stacktrace but not leak information)
report_errors = False
#: if show error dialog with stacktrace, if not then only print it to console
show_error_dialog = False
#: if load inner plugins (plugins that are not in separate package)
custom_plugin_load = False
#: if check for updates on startup,
check_for_updates = True
#: if run in develop mode (show developer tab in advanced window)
develop = False

save_suffix = ""
#: path to folder where save settings
save_folder = os.path.join(
    os.environ.get("PARTSEG_SETTINGS_DIR", appdirs.user_data_dir(APP_NAME, APP_LAB)),
    str(packaging.version.parse(__version__).base_version),
)

sentry_url = os.environ.get("PARTSEG_SENTRY_URL", "https://d4118280b73d4ee3a0222d0b17637687@sentry.io/1309302")

__all__ = (
    "always_report",
    "auto_report",
    "check_for_updates",
    "custom_plugin_load",
    "develop",
    "report_errors",
    "save_folder",
    "save_suffix",
    "sentry_url",
    "show_error_dialog",
)



import os
import re
from importlib.metadata import requires, version

import pytest

requires_list = [x for x in requires("PartSeg") if ";" not in x and ">=" in x]
requires_regexp = re.compile(r"^([A-Za-z][A-Za-z0-9\-]+).*>=(\d[^,)]+)")


@pytest.mark.skipif(not os.environ.get("MINIMAL_REQUIREMENTS", ""), reason="not running minimum requirement test")
@pytest.mark.parametrize("requirement_specifier", requires_list)
def test_min_req(requirement_specifier):
    match = requires_regexp.match(requirement_specifier)
    assert match is not None
    package_name = match.group(1)
    version_specifier = match.group(2)
    assert version(package_name) == version_specifier



import os
from collections import Counter, defaultdict
from functools import partial
from pathlib import Path

from qtpy.QtCore import Qt, QTimer, Signal, Slot
from qtpy.QtGui import QFontMetrics, QMouseEvent, QResizeEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QAction,
    QApplication,
    QCheckBox,
    QDialog,
    QFileDialog,
    QGridLayout,
    QInputDialog,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QMessageBox,
    QPushButton,
    QTreeWidget,
    QTreeWidgetItem,
    QWidget,
)

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.custom_load_dialog import CustomLoadDialog, LoadProperty
from PartSeg.common_gui.exception_hooks import load_data_exception_hook
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore.io_utils import LoadBase
from PartSegCore.project_info import ProjectInfoBase


class MultipleFilesTreeWidget(QTreeWidget):
    context_load = Signal(QTreeWidgetItem)
    context_compare = Signal(QTreeWidgetItem)
    context_forget = Signal(QTreeWidgetItem)

    def __init__(self, compare, parent=None):
        super().__init__(parent)
        self.compare = compare
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

    def showContextMenu(self, point):
        item = self.itemAt(point)
        if item is None:
            return
        menu = QMenu()
        if item.parent() is not None:
            action1 = QAction("Load")
            action1.triggered.connect(partial(self.context_load.emit, item))
            menu.addAction(action1)
            if self.compare and item.text(0) not in ["raw image", "image with mask"]:
                action2 = QAction("Compare")
                action2.triggered.connect(partial(self.context_compare.emit, item))
                menu.addAction(action2)
        action = QAction("Forget")
        action.triggered.connect(partial(self.context_forget.emit, item))
        menu.addAction(action)
        menu.exec_(self.mapToGlobal(point))

    def set_show_compare(self, compare: bool):
        self.compare = compare

    def mouseMoveEvent(self, event):  # pylint: disable=no-self-use
        QApplication.setOverrideCursor(Qt.CursorShape.ArrowCursor)
        super().mouseMoveEvent(event)


class LoadRecentFiles(QDialog):
    def __init__(self, settings: BaseSettings, parent=None):
        super().__init__(parent)
        self.settings = settings
        self.file_list = QListWidget()
        self.file_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.cancel_btn = QPushButton("Cancel", clicked=self.reject)
        self.load_btn = QPushButton("Load", clicked=self.accept)

        for name_list, method in settings.get_last_files_multiple():
            entry = f"{name_list[0]} {method}"
            item = QListWidgetItem(entry, self.file_list)
            item.setData(Qt.ItemDataRole.UserRole, (name_list, method))

        last_set = {(tuple(x), y) for x, y in settings.get_last_files_multiple()}
        for name_list, method in settings.get_last_files():
            if (tuple(name_list), method) in last_set:
                continue
            entry = f"{name_list[0]} {method}"
            item = QListWidgetItem(entry, self.file_list)
            item.setData(Qt.ItemDataRole.UserRole, (name_list, method))

        layout = QGridLayout()
        layout.addWidget(QLabel("Select files"))
        layout.addWidget(self.file_list, 1, 0, 1, 2)
        layout.addWidget(self.cancel_btn, 2, 0)
        layout.addWidget(self.load_btn, 2, 1)

        self.setLayout(layout)
        self.resize(
            *self.settings.get_from_profile("multiple_files_dialog_size", (self.size().width(), self.size().height()))
        )

    def get_files(self) -> list[tuple[list[str], str]]:
        return [item.data(Qt.ItemDataRole.UserRole) for item in self.file_list.selectedItems()]

    def accept(self) -> None:
        self.settings.set_in_profile("multiple_files_dialog_size", (self.size().width(), self.size().height()))
        super().accept()


class MultipleFileWidget(QWidget):
    _add_state = Signal(object, bool)

    def __init__(self, settings: BaseSettings, load_dict: dict[str, LoadBase], compare_in_context_menu=False):
        super().__init__()
        self.settings = settings
        self.state_dict: dict[str, dict[str, ProjectInfoBase]] = defaultdict(dict)
        self.state_dict_count = Counter()
        self.file_list = []
        self.load_register = load_dict
        self.file_view = MultipleFilesTreeWidget(compare_in_context_menu)
        self.file_view.header().close()
        self.save_state_btn = QPushButton("Save state")
        self.save_state_btn.setStyleSheet("QPushButton{font-weight: bold;}")
        self.load_files_btn = QPushButton("Load Files")
        self.load_recent_files_btn = QPushButton("Load recent Files")
        self.forget_btn = QPushButton("Forget")
        self.forget_all_btn = QPushButton("Forget all")

        self.save_state_btn.clicked.connect(self.save_state)
        self.forget_btn.clicked.connect(self.forget)
        self.load_files_btn.clicked.connect(self.load_files)
        self.load_recent_files_btn.clicked.connect(self.load_recent)
        self.file_view.itemDoubleClicked.connect(self.load_state)
        self.file_view.context_load.connect(self.load_state)
        self.forget_all_btn.clicked.connect(self.forget_all)
        self.last_point = None

        self.custom_names_chk = QCheckBox("Custom names")

        layout = QGridLayout()
        layout.addWidget(self.file_view, 0, 0, 1, 2)
        layout.addWidget(self.save_state_btn, 1, 0)
        layout.addWidget(self.load_files_btn, 1, 1)
        layout.addWidget(self.load_recent_files_btn, 2, 1)
        layout.addWidget(self.forget_btn, 2, 0)
        layout.addWidget(self.forget_all_btn, 3, 0)
        layout.addWidget(self.custom_names_chk, 3, 1)

        self.setLayout(layout)
        self.setMouseTracking(True)
        self.file_view.setMouseTracking(True)
        self.file_view.context_load.connect(self.load_state)
        self.file_view.context_compare.connect(self.load_compare)
        self.file_view.context_forget.connect(self.forget_action)
        self.error_list = []

        self._add_state.connect(self.save_state_action)
        self.settings.connect_("multiple_files_widget", self.view_changed)
        self.view_changed()

    def load_recent(self):
        dial = LoadRecentFiles(self.settings, self)
        if not dial.exec_():
            return

        dial_fun = ExecuteFunctionDialog(
            self.load_recent_fun, [dial.get_files()], exception_hook=load_data_exception_hook
        )
        dial_fun.exec_()

    def load_recent_fun(self, load_list, range_changed, step_changed):
        range_changed(0, len(load_list))
        for i, (file_list, method) in enumerate(load_list):
            load_class = self.load_register[method]
            state: ProjectInfoBase = load_class.load(file_list)
            self._add_state.emit(state, False)
            step_changed(i)
        for file_list, method in reversed(load_list):
            self.settings.add_last_files_multiple(file_list, method)

    def view_changed(self):
        self.setVisible(self.settings.get("multiple_files_widget", False))

    def execute_load_files(self, load_data: LoadProperty, range_changed, step_changed):
        range_changed(0, len(load_data.load_location))
        loaded_list = []
        for i, el in enumerate(load_data.load_location, 1):
            load_list = [el]
            while load_data.load_class.number_of_files() > len(load_list):
                load_list.append(load_data.load_class.get_next_file(load_list))
                if not os.path.exists(load_list[-1]):
                    self.error_list.append(el)
                    step_changed(i)
                    continue
            state: ProjectInfoBase = load_data.load_class.load(load_list)
            loaded_list.append((load_list, load_data.load_class.get_name()))
            self._add_state.emit(state, False)
            step_changed(i)

        for el in reversed(loaded_list):
            self.settings.add_last_files_multiple(*el)

    def load_files(self):
        dial = MultipleLoadDialog(self.load_register, self.settings.get_path_history())
        dial.setDirectory(self.settings.get("io.multiple_open_directory", str(Path.home())))
        dial.selectNameFilter(self.settings.get("io.multiple_open_filter", next(iter(self.load_register.keys()))))
        self.error_list = []
        if dial.exec_():
            result = dial.get_result()
            load_dir = os.path.dirname(result.load_location[0])
            self.settings.set("io.multiple_open_directory", load_dir)
            self.settings.add_path_history(load_dir)
            self.settings.set("io.multiple_open_filter", result.selected_filter)

            dial_fun = ExecuteFunctionDialog(self.execute_load_files, [result], exception_hook=load_data_exception_hook)
            dial_fun.exec_()
            if self.error_list:
                errors_message = QMessageBox()
                errors_message.setText("There are errors during load files")
                errors_message.setInformativeText("During load files cannot found some of files on disc")
                errors_message.setStandardButtons(QMessageBox.Ok)
                text = "\n".join(f"File: {x[0]}\n{x[1]}" for x in self.error_list)
                errors_message.setDetailedText(text)
                errors_message.exec_()

    def load_state(self, item, _column=1):
        if item.parent() is None:
            return
        file_name = self.file_list[self.file_view.indexOfTopLevelItem(item.parent())]
        state_name = item.text(0)
        project_info = self.state_dict[file_name][state_name]
        try:
            self.parent().parent().parent().set_data(project_info)
        except AttributeError:
            self.settings.set_project_info(project_info)

    def load_compare(self, item):
        if item.parent() is None:
            return
        file_name = self.file_list[self.file_view.indexOfTopLevelItem(item.parent())]
        if self.settings.image.file_path != file_name:
            QMessageBox.information(self, "Wrong file", "Please select same file as main")
            return
        state_name = item.text(0)
        project_info = self.state_dict[file_name][state_name]
        if hasattr(self.settings, "set_segmentation_to_compare"):
            self.settings.set_segmentation_to_compare(project_info.roi_info)

    def save_state(self):
        state: ProjectInfoBase = self.settings.get_project_info()
        custom_name = self.custom_names_chk.isChecked()
        self.save_state_action(state, custom_name)

    def save_state_action(self, state: ProjectInfoBase, custom_name):
        # TODO left ellipsis
        if isinstance(state, list):
            self.add_states(state)
            return
        if not isinstance(state, ProjectInfoBase):  # workaround for PointsInfo load
            return
        normed_file_path = os.path.normpath(state.file_path)
        sub_dict = self.state_dict[normed_file_path]
        name = f"state {self.state_dict_count[normed_file_path]+1}"
        if custom_name:
            name, ok = QInputDialog.getText(self, "Save name", "Save name:", text=name)
            if not ok:
                return
            while name in sub_dict or name in ["raw image", "image with mask"]:
                name, ok = QInputDialog.getText(self, "Save name", "Save name (previous in use):", text=name)
                if not ok:
                    return
        try:
            index = self.file_list.index(os.path.normpath(normed_file_path))
            item = self.file_view.topLevelItem(index)
        except ValueError:
            metric = QFontMetrics(self.file_view.font())
            width = self.file_view.width() - 45
            clipped_text = metric.elidedText(normed_file_path, Qt.TextElideMode.ElideLeft, width)
            item = QTreeWidgetItem(self.file_view, [clipped_text])
            item.setToolTip(0, normed_file_path)
            self.file_list.append(normed_file_path)
            QTreeWidgetItem(item, ["raw image"])
            sub_dict["raw image"] = state.get_raw_copy()
            if state.is_masked():
                QTreeWidgetItem(item, ["image with mask"])
                sub_dict["image with mask"] = state.get_raw_mask_copy()

        item.setExpanded(True)
        if state.is_raw():
            return
        it = QTreeWidgetItem(item, [name])
        self.file_view.setCurrentItem(it)
        sub_dict[name] = state
        self.state_dict_count[state.file_path] += 1

    def forget(self):
        if not self.forget_btn.isEnabled():
            return
        self.forget_btn.setDisabled(True)
        item: QTreeWidgetItem = self.file_view.currentItem()
        self.forget_action(item)
        QTimer().singleShot(500, self.enable_forget)

    def forget_all(self):
        for index in range(self.file_view.topLevelItemCount(), -1, -1):
            self.forget_action(self.file_view.topLevelItem(index))

    def forget_action(self, item):
        if item is None:
            return
        if isinstance(item.parent(), QTreeWidgetItem):
            index = self.file_view.indexOfTopLevelItem(item.parent())
            text = self.file_list[index]
            if item.text(0) not in self.state_dict[text]:
                return
            del self.state_dict[text][item.text(0)]
            parent = item.parent()
            parent.removeChild(item)
            if parent.childCount() == 0:
                self.file_view.takeTopLevelItem(index)
                self.file_list.remove(text)

        else:
            index = self.file_view.indexOfTopLevelItem(item)
            text = self.file_list[index]
            del self.state_dict[text]
            del self.state_dict_count[text]
            self.file_list.remove(text)
            self.file_view.takeTopLevelItem(index)

    @Slot()
    def enable_forget(self):
        self.forget_btn.setEnabled(True)

    def resizeEvent(self, event: QResizeEvent):
        metric = QFontMetrics(self.file_view.font())
        width = self.file_view.width() - 45
        for i, text in enumerate(self.file_list):
            clipped_text = metric.elidedText(text, Qt.TextElideMode.ElideLeft, width)
            item: QTreeWidgetItem = self.file_view.topLevelItem(i)
            item.setText(0, clipped_text)

    def mousePressEvent(self, event: QMouseEvent):
        if event.x() > self.width() - 20:
            self.last_point = event.pos()

    def mouseMoveEvent(self, event: QMouseEvent):
        if event.x() > self.width() - 20:
            QApplication.setOverrideCursor(Qt.CursorShape.SplitHCursor)
        else:
            QApplication.setOverrideCursor(Qt.CursorShape.ArrowCursor)
        if self.last_point is None or not (event.buttons() & Qt.MouseButton.LeftButton):
            return
        new_width = event.x() + 10
        new_width = max(new_width, 150)
        new_width = min(new_width, 600)

        self.setMinimumWidth(new_width)

    def leaveEvent(self, _):  # pylint: disable=no-self-use
        QApplication.setOverrideCursor(Qt.CursorShape.ArrowCursor)

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.last_point = None

    def set_compare_in_context_menu(self, compare: bool):
        self.file_view.set_show_compare(compare)

    def add_states(self, states: list[ProjectInfoBase]):
        """add multiple states to widget"""
        for el in states:
            self.save_state_action(el, False)


class MultipleLoadDialog(CustomLoadDialog):
    def __init__(self, load_register, history=None):
        load_register = {key: val for key, val in load_register.items() if not val.partial()}
        super().__init__(load_register=load_register, history=history)
        self.setFileMode(QFileDialog.FileMode.ExistingFiles)

    def accept(self):
        self.files_list.extend(self.selectedFiles())
        QFileDialog.accept(self)



import os
import typing
from importlib.metadata import version

import numpy as np
from packaging.version import parse as parse_version

from PartSeg.plugins.napari_widgets._settings import get_settings
from PartSegCore import UNIT_SCALE
from PartSegCore.analysis import ProjectTuple
from PartSegCore.io_utils import LoadBase, WrongFileTypeException
from PartSegCore.mask.io_functions import MaskProjectTuple
from PartSegCore.universal_const import format_layer_name
from PartSegImage import Image


@typing.overload
def adjust_color(color: str) -> str: ...


@typing.overload
def adjust_color(color: list[int]) -> list[float]: ...


def adjust_color(color: typing.Union[str, list[int]]) -> typing.Union[str, tuple[float]]:
    # as napari ignore alpha channel in color, and adding it to
    # color cause that napari fails to detect that such colormap is already present
    # in this function I remove alpha channel if it is present
    if isinstance(color, str) and color.startswith("#"):
        if len(color) == 9:
            # case when color is in format #RRGGBBAA
            return color[:7]
        if len(color) == 5:
            # case when color is in format #RGBA
            return color[:4]
    elif isinstance(color, list):
        return (color[i] / 255 for i in range(3))
    # If not fit to an earlier case, return as is.
    # Maybe napari will handle it
    return color


if parse_version(version("napari")) >= parse_version("0.4.19a1"):

    def add_color(image: Image, idx: int) -> dict:
        return {
            "colormap": adjust_color(image.get_colors()[idx]),
        }

else:

    def add_color(image: Image, idx: int) -> dict:  # noqa: ARG001
        # Do nothing, as napari is not able to pass hex color to image
        # the image and idx are present to keep the same signature
        return {}


def _image_to_layers(project_info, scale, translate):
    settings = get_settings()
    filename = os.path.basename(project_info.file_path)
    res_layers = []
    if project_info.image.name == "ROI" and project_info.image.channels == 1:
        res_layers.append(
            (
                project_info.image.get_channel(0),
                {
                    "scale": scale,
                    "name": format_layer_name(
                        settings.layer_naming_format, filename, project_info.image.channel_names[0]
                    ),
                    "translate": translate,
                },
                "labels",
            )
        )
    else:
        res_layers.extend(
            (
                project_info.image.get_channel(i),
                {
                    "scale": scale,
                    "name": format_layer_name(
                        settings.layer_naming_format, filename, project_info.image.channel_names[i]
                    ),
                    "blending": "additive",
                    "translate": translate,
                    "metadata": project_info.image.metadata,
                    **add_color(project_info.image, i),
                },
                "image",
            )
            for i in range(project_info.image.channels)
        )

    return res_layers


def project_to_layers(project_info: typing.Union[ProjectTuple, MaskProjectTuple]):
    res_layers = []
    if project_info.image is not None and not isinstance(project_info.image, str):
        settings = get_settings()
        scale = project_info.image.normalized_scaling(UNIT_SCALE[settings.io_units.value])
        translate = project_info.image.shift
        translate = (0,) * (len(project_info.image.axis_order.replace("C", "")) - len(translate)) + translate
        res_layers.extend(_image_to_layers(project_info, scale, translate))
        if project_info.roi_info.roi is not None:
            res_layers.append(
                (
                    project_info.image.fit_array_to_image(project_info.roi_info.roi),
                    {"scale": scale, "name": "ROI", "translate": translate},
                    "labels",
                )
            )
        if project_info.roi_info.alternative:
            res_layers.extend(
                (
                    project_info.image.fit_array_to_image(roi),
                    {
                        "scale": scale,
                        "name": name,
                        "translate": translate,
                        "visible": False,
                    },
                    "labels",
                )
                for name, roi in project_info.roi_info.alternative.items()
            )

        if project_info.mask is not None:
            res_layers.append(
                (
                    project_info.image.fit_array_to_image(project_info.mask),
                    {"scale": scale, "name": "Mask", "translate": translate},
                    "labels",
                )
            )
    else:
        if isinstance(project_info, MaskProjectTuple) and project_info.spacing is not None:
            scale = np.multiply(project_info.spacing, 10**9)
        else:
            scale = None
        if project_info.roi_info.roi is not None:
            res_layers.append(
                (
                    project_info.roi_info.roi,
                    {"scale": scale, "name": "ROI"},
                    "labels",
                )
            )
    return res_layers


def partseg_loader(loader: type[LoadBase], path: str):
    load_locations = [path]
    load_locations.extend(loader.get_next_file(load_locations) for _ in range(1, loader.number_of_files()))

    try:
        project_info = loader.load(load_locations)
    except WrongFileTypeException:  # pragma: no cover
        return None

    if isinstance(project_info, (ProjectTuple, MaskProjectTuple)):
        return project_to_layers(project_info)
    return None  # pragma: no cover



from contextlib import suppress
from functools import partial
from typing import Union

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QCheckBox, QDialog, QHBoxLayout, QLabel, QPushButton, QSpinBox, QVBoxLayout, QWidget
from superqt import QEnumComboBox

from PartSeg.common_backend.base_settings import BaseSettings, ImageSettings
from PartSegCore.image_operations import RadiusType
from PartSegCore.mask_create import MaskProperty
from PartSegCore.segmentation.algorithm_base import calculate_operation_radius


def off_widget(widget: QWidget, combo_box: QEnumComboBox, _val=None):
    widget.setDisabled(combo_box.currentEnum() == RadiusType.NO)


class MaskWidget(QWidget):
    values_changed = Signal()

    def __init__(self, settings: ImageSettings, parent=None):
        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        self.settings = settings
        self.dilate_radius = QSpinBox()
        self.dilate_radius.setRange(-100, 100)
        self.dilate_radius.setSingleStep(1)
        self.dilate_radius.setDisabled(True)
        self.dilate_dim = QEnumComboBox(enum_class=RadiusType)
        self.dilate_dim.setToolTip("With minus radius mask will be eroded")
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(partial(off_widget, self.dilate_radius, self.dilate_dim))

        self.radius_information = QLabel()

        # noinspection PyUnresolvedReferences
        self.dilate_radius.valueChanged.connect(self.dilate_change)
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(self.dilate_change)

        self.fill_holes = QEnumComboBox(enum_class=RadiusType)
        self.max_holes_size = QSpinBox()
        self.max_holes_size.setRange(-1, 10000)
        self.max_holes_size.setValue(-1)
        self.max_holes_size.setSingleStep(100)
        self.max_holes_size.setDisabled(True)
        self.max_holes_size.setToolTip("Maximum size of holes to be closed. -1 means that all holes will be closed")
        # noinspection PyUnresolvedReferences
        self.fill_holes.currentIndexChanged.connect(partial(off_widget, self.max_holes_size, self.fill_holes))

        self.save_components = QCheckBox()
        self.clip_to_mask = QCheckBox()
        self.reversed_mask = QCheckBox()

        # noinspection PyUnresolvedReferences
        self.dilate_radius.valueChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.dilate_dim.currentIndexChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.fill_holes.currentIndexChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.max_holes_size.valueChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.save_components.stateChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.clip_to_mask.stateChanged.connect(self._value_changed_wrap)
        # noinspection PyUnresolvedReferences
        self.reversed_mask.stateChanged.connect(self._value_changed_wrap)
        self.setup_ui()

    def setup_ui(self):
        layout = QVBoxLayout()
        layout1 = QHBoxLayout()
        layout1.addWidget(QLabel("Dilate mask:"))
        layout1.addWidget(self.dilate_dim)
        layout1.addWidget(QLabel("radius (in pix):"))
        layout1.addWidget(self.dilate_radius)
        layout.addLayout(layout1)
        layout2 = QHBoxLayout()
        layout2.addWidget(QLabel("Fill holes:"))
        layout2.addWidget(self.fill_holes)
        layout2.addWidget(QLabel("Max size:"))
        layout2.addWidget(self.max_holes_size)
        layout.addLayout(layout2)
        layout3 = QHBoxLayout()
        comp_lab = QLabel("Save components:")
        comp_lab.setToolTip(
            "save components information in mask. Dilation, holes filing will be done separately for each component"
        )
        self.save_components.setToolTip(comp_lab.toolTip())
        layout3.addWidget(comp_lab)
        layout3.addWidget(self.save_components)
        layout3.addStretch()
        clip_mask = QLabel("Clip to previous mask:")
        clip_mask.setToolTip("Useful dilated new mask")
        layout3.addWidget(clip_mask)
        layout3.addWidget(self.clip_to_mask)
        layout.addLayout(layout3)
        layout4 = QHBoxLayout()
        layout4.addWidget(QLabel("Reversed mask:"))
        layout4.addWidget(self.reversed_mask)
        layout4.addStretch(1)
        layout.addLayout(layout4)
        self.setLayout(layout)
        self.dilate_change()

    def _value_changed_wrap(self, _val=None):
        # noinspection PyUnresolvedReferences
        self.values_changed.emit()

    def get_dilate_radius(self) -> Union[int, list[int]]:
        radius = calculate_operation_radius(
            self.dilate_radius.value(), self.settings.image_spacing, self.dilate_dim.currentEnum()
        )
        if isinstance(radius, (list, tuple)):
            return [int(x + 0.5) for x in radius]
        return int(radius)

    def dilate_change(self):
        if self.dilate_radius.value() == 0 or self.dilate_dim.currentEnum() == RadiusType.NO:
            self.radius_information.setText("Dilation radius: 0")
        else:
            dilate_radius = self.get_dilate_radius()
            if isinstance(dilate_radius, list):
                self.radius_information.setText(f"Dilation radius: {dilate_radius[::-1]}")
            else:
                self.radius_information.setText(f"Dilation radius: {dilate_radius}")

    def get_mask_property(self):
        return MaskProperty(
            dilate=self.dilate_dim.currentEnum() if self.dilate_radius.value() != 0 else RadiusType.NO,
            dilate_radius=self.dilate_radius.value() if self.dilate_dim.currentEnum() != RadiusType.NO else 0,
            fill_holes=self.fill_holes.currentEnum() if self.max_holes_size.value() != 0 else RadiusType.NO,
            max_holes_size=self.max_holes_size.value() if self.fill_holes.currentEnum() != RadiusType.NO else 0,
            save_components=self.save_components.isChecked(),
            clip_to_mask=self.clip_to_mask.isChecked(),
            reversed_mask=self.reversed_mask.isChecked(),
        )

    def set_mask_property(self, prop: MaskProperty):
        self.dilate_dim.setCurrentEnum(prop.dilate)
        self.dilate_radius.setValue(prop.dilate_radius)
        self.fill_holes.setCurrentEnum(prop.fill_holes)
        self.max_holes_size.setValue(prop.max_holes_size)
        self.save_components.setChecked(prop.save_components)
        self.clip_to_mask.setChecked(prop.clip_to_mask)
        self.reversed_mask.setChecked(prop.reversed_mask)


class MaskDialogBase(QDialog):
    def __init__(self, settings: BaseSettings):
        super().__init__()
        self.setWindowTitle("Mask manager")
        self.settings = settings
        main_layout = QVBoxLayout()
        self.mask_widget = MaskWidget(settings, self)
        main_layout.addWidget(self.mask_widget)
        with suppress(KeyError):
            mask_property = self.settings.get("mask_manager.mask_property")
            self.mask_widget.set_mask_property(mask_property)

        self.reset_next_btn = QPushButton("Reset Next")
        self.reset_next_btn.clicked.connect(self.reset_next_fun)
        if settings.history_redo_size() == 0:
            self.reset_next_btn.setDisabled(True)
        self.set_next_btn = QPushButton("Set Next")
        if settings.history_redo_size() == 0:
            self.set_next_btn.setDisabled(True)
        self.set_next_btn.clicked.connect(self.set_next)
        self.cancel = QPushButton("Cancel", self)
        self.cancel.clicked.connect(self.close)
        self.prev_button = QPushButton(f"Previous mask ({settings.history_size()})", self)
        if settings.history_size() == 0:
            self.prev_button.setDisabled(True)
        self.next_button = QPushButton(f"Next mask ({settings.history_redo_size()})", self)
        if settings.history_redo_size() == 0:
            self.next_button.setText("Next mask (new)")
        self.next_button.clicked.connect(self.next_mask)
        self.prev_button.clicked.connect(self.prev_mask)
        op_layout = QHBoxLayout()
        op_layout.addWidget(self.mask_widget.radius_information)
        main_layout.addLayout(op_layout)
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.cancel)
        button_layout.addWidget(self.set_next_btn)
        button_layout.addWidget(self.reset_next_btn)
        main_layout.addLayout(button_layout)
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.prev_button)
        button_layout.addWidget(self.next_button)
        main_layout.addLayout(button_layout)
        self.setLayout(main_layout)
        if self.settings.history_redo_size():
            mask_prop: MaskProperty = self.settings.history_next_element().mask_property
            self.mask_widget.set_mask_property(mask_prop)
        self.mask_widget.values_changed.connect(self.values_changed)

    def set_next(self):
        if self.settings.history_redo_size():
            self.mask_widget.set_mask_property(self.settings.history_next_element().mask_property)

    def values_changed(self):
        if (
            self.settings.history_redo_size()
            and self.mask_widget.get_mask_property() == self.settings.history_next_element().mask_property
        ):
            self.next_button.setText(f"Next mask ({self.settings.history_redo_size()})")
        else:
            self.next_button.setText("Next mask (new)")

    def reset_next_fun(self):
        self.settings.history_redo_clean()
        self.next_button.setText("Next mask (new)")
        self.reset_next_btn.setDisabled(True)

    def next_mask(self):
        raise NotImplementedError

    def prev_mask(self):
        raise NotImplementedError



import functools

from PartSeg.plugins.napari_io.loader import partseg_loader
from PartSegCore.analysis.load_functions import LoadStackImage


def napari_get_reader(path: str):
    return next(
        (
            functools.partial(partseg_loader, LoadStackImage)
            for extension in LoadStackImage.get_extensions()
            if path.lower().endswith(extension.lower())
        ),
        None,
    )



import typing

import numpy as np
from napari.utils import Colormap
from napari.utils.colormaps import make_colorbar
from qtpy.QtGui import QImage

ColorMapDict = typing.MutableMapping[str, tuple[Colormap, bool]]


class NumpyQImage(QImage):
    """
    Class for fix problem with PySide2 QImage implementation (non copied buffer)
    """

    def __init__(self, image: np.ndarray):
        super().__init__(
            image.data,
            image.shape[1],
            image.shape[0],
            image.dtype.itemsize * image.shape[1] * image.shape[2],
            QImage.Format.Format_RGBA8888,
        )
        self.image = image


def convert_colormap_to_image(colormap: Colormap) -> NumpyQImage:
    """
    convert colormap to image of size (512, 1)

    :param colormap: colormap to convert
    :return: Color Bar image
    """
    return NumpyQImage(np.array(make_colorbar(colormap, size=(1, 512))))


def create_colormap_image(colormap: str, color_dict: ColorMapDict) -> NumpyQImage:
    """
    Convert named colormap to image of size (512, 1)

    :param colormap: must be key in color_dict
    :param color_dict: dict mapping name to definition of colormap
    :return: Color Bar image
    """
    return convert_colormap_to_image(color_dict[colormap][0])



from PartSeg.plugins.modeling_save.save_modeling_data import SaveModeling


def register():
    from PartSeg import state_store

    if state_store.custom_plugin_load:
        from PartSegCore.register import RegisterEnum
        from PartSegCore.register import register as register_fun

        register_fun(SaveModeling, RegisterEnum.analysis_save)



from enum import Enum, auto
from typing import Callable, Optional, Union

import numpy as np

from PartSegCore.algorithm_describe_base import AlgorithmProperty
from PartSegCore.image_transforming.transform_base import TransformBase
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


class CombineMode(Enum):
    Max = auto()
    Sum = auto()


class CombineChannels(TransformBase):
    @classmethod
    def get_fields(cls):
        return [AlgorithmProperty("combine_mode", "Combine Mode", CombineMode.Sum)]

    @classmethod
    def get_fields_per_dimension(cls, image: Image) -> list[Union[str, AlgorithmProperty]]:
        return [
            AlgorithmProperty("combine_mode", "Combine Mode", CombineMode.Sum),
            *[AlgorithmProperty(f"channel_{i}", f"Channel {i}", False) for i in range(image.channels)],
        ]

    @classmethod
    def get_name(cls):
        return "Combine channels"

    @classmethod
    def transform(
        cls,
        image: Image,
        roi_info: Optional[ROIInfo],
        arguments: dict,
        callback_function: Optional[Callable[[str, int], None]] = None,
    ) -> tuple[Image, Optional[ROIInfo]]:
        channels = [i for i, x in enumerate(x for x in arguments.items() if x[0].startswith("channel")) if x[1]]
        if not channels:
            return image, roi_info
        channel_array = [image.get_channel(i) for i in channels]
        if len(channels) == 1:
            new_channel = channel_array[0]
        elif arguments["combine_mode"] == CombineMode.Max:
            new_channel = np.max(channel_array, axis=0)
        else:
            new_channel = np.sum(channel_array, axis=0)
        all_channels = [image.get_channel(i) for i in range(image.channels)]
        all_channels.append(new_channel)
        channel_names = [*image.channel_names, "combined"]
        contrast_limits = [*image.get_ranges(), (np.min(new_channel), np.max(new_channel))]
        return image.substitute(data=all_channels, channel_names=channel_names, ranges=contrast_limits), roi_info

    @classmethod
    def calculate_initial(cls, image: Image):
        min_val = min(image.spacing)
        return {
            f"scale_{letter}": x / min_val for x, letter in zip(image.spacing, image.get_dimension_letters().lower())
        }



import collections
import importlib
import inspect
import itertools
import json
import pathlib
import pprint
import sys
import typing
import warnings
from contextlib import suppress
from enum import Enum, EnumMeta

# TODO read about dataclsss and maybe apply

_class_template = """\
from builtins import property as _property, tuple as _tuple
from operator import attrgetter as _attrgetter
from collections import OrderedDict
import typing
\"\"\"
{imports}
\"\"\"

class {typename}({base_classes}):
    "{typename}({signature})"
    __slots__ = {slots!r}

    _fields = {field_names!r}
    _root = True

    #def __new__(cls, {signature}):
    #    ob = super().__new__(cls)
    #    cls.__init__(ob, {arg_list})
    #    return ob

    def __init__(self, {signature}):
        'Create new instance of {typename}({arg_list})'
        {init_content}

    @classmethod
    def _make(cls, iterable):
        'Make a new {typename} object from a sequence or iterable'
        return cls.make_(iterable)

    @classmethod
    def make_(cls, iterable):
        'Make a new {typename} object from a sequence or iterable'
        if isinstance(iterable, dict):
            result = cls(**iterable)
        else:
            result = cls(*iterable)
        return result

    def replace_(self, **kwargs):
        'Return a new {typename} object replacing specified fields with new values'
        return self._replace(**kwargs)

    def _replace(self, **kwargs):
        dkt = self.asdict()
        dkt.update(kwargs)
        return self.__class__(**dkt)

    def __repr__(self):
        'Return a nicely formatted representation string'
        return self.__class__.__name__ + '({repr_fmt})' % _attrgetter(*self.__slots__)(self)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.as_tuple() == other.as_tuple()

    def asdict(self):
        'Return a new OrderedDict which maps field names to their values.'
        return OrderedDict(zip(self._fields, _attrgetter(*self.__slots__)(self)))

    def as_dict(self):
        'Return a new OrderedDict which maps field names to their values.'
        return dict(zip(self._fields, _attrgetter(*self.__slots__)(self)))

    def as_tuple(self):
        return {tuple_fields}

    def __getnewargs__(self):
        'Return self as a plain tuple.  Used by copy and pickle.'
        return tuple(_attrgetter(*self.__slots__)(self))

    #def __setattr__(self, name, value):
    #    raise AttributeError('can\\\'t set attribute')

{field_definitions}
"""

_repr_template = "{name}=%r"

_field_template = """\
    {name} = _property(_attrgetter("_{name}"), doc='getter for attribute {name}')
"""

T = typing.TypeVar("T")


class RegisterClass(typing.Generic[T]):
    def __init__(self):
        self.exact_class_register = {}
        self.predict_class_register = collections.defaultdict(list)

    def register_class(self, cls: T, old_name=None):
        path = extract_type_info(cls)[0]
        name = cls.__name__
        if path in self.exact_class_register:
            raise ValueError("class already registered")
        self.exact_class_register[path] = cls
        self.predict_class_register[name].append(cls)
        if old_name is not None:
            if isinstance(old_name, str):
                self.predict_class_register[old_name].append(cls)
            elif isinstance(old_name, typing.Iterable):
                for ol in old_name:
                    self.predict_class_register[ol].append(cls)

    def get_class(self, path: str) -> T:
        if path in self.exact_class_register:
            return self.exact_class_register[path]
        name = path[path.rfind(".") + 1 :]
        if name not in self.predict_class_register:
            with suppress(ImportError):
                importlib.import_module(path[: path.rfind(".")])
        if name in self.predict_class_register:
            if len(self.predict_class_register[name]) == 1:
                return self.predict_class_register[name][0]
            return iter(self.predict_class_register[name])
        raise ValueError(f"unregistered class {path}")

    def clear(self):  # for testing purpose
        self.exact_class_register.clear()
        self.predict_class_register.clear()


base_serialize_register = RegisterClass()
enum_register: RegisterClass[EnumMeta] = RegisterClass()


def extract_type_name(type_):
    if not hasattr(type_, "__name__"):
        if hasattr(type_, "__module__") and type_.__module__ == "typing" and hasattr(type_, "__origin__"):
            return type_.__origin__
        return str(type_)
    return type_.__name__


def extract_type_info(type_):
    # noinspection PyUnresolvedReferences
    if not hasattr(type_, "__module__"):
        return type_.__name__, None

    if type_.__module__ == "typing":
        return str(type_), type_.__module__
    return f"{type_.__module__}.{type_.__name__}", type_.__module__


_prohibited = (
    "__new__",
    "__init__",
    "__slots__",
    "__getnewargs__",
    "_fields",
    "_field_defaults",
    "_field_types",
    "_make",
    "replace_",
    "asdict",
    "_source",
    "asdict",
    "__setattr__",
)

_special = ("__module__", "__name__", "__qualname__", "__annotations__", "_reloading")

omit_list = (typing.TypeVar,)


def add_classes(types_list, translate_dict, global_state):
    ll = list(types_list)
    for type_ in ll:
        if type_ in translate_dict or isinstance(type_, omit_list):
            continue
        if hasattr(type_, "__module__") and type_.__module__ == "typing":
            if (
                hasattr(type_, "__args__")
                and isinstance(type_.__args__, collections.abc.Iterable)
                and (sub_types := [x for x in type_.__args__ if not isinstance(x, omit_list)])
            ):
                add_classes(sub_types, translate_dict, global_state)
                if type_._name is None:  # pylint: disable=protected-access
                    type_str = str(type_.__origin__)
                else:
                    type_str = f"typing.{type_._name!s}"  # pylint: disable=protected-access
                type_str += "[" + ", ".join(translate_dict[x] for x in sub_types) + "]"
                translate_dict[type_] = type_str
                continue

            if isinstance(type_, typing.ForwardRef):
                translate_dict[type_] = f"'{type_.__forward_arg__}'"
                continue
            translate_dict[type_] = str(type_)
            continue

        name = extract_type_name(type_)
        while name in global_state:
            name += "a"
        translate_dict[type_] = name

        global_state[name] = type_


def _make_class(typename, types, defaults_dict, base_classes, readonly):
    if base_classes:
        # TODO add function inheritance
        types_ = types
        types = {}
        for el in base_classes:
            if hasattr(el, "__annotations__"):
                types.update(el.__annotations__)
        types.update(types_)
    field_names = list(types.keys())
    import_set = set()
    type_dict = {}
    for name_, type_ in types.items():
        type_str, module = extract_type_info(type_)
        type_dict[name_] = type_str
        if module:
            import_set.add(f"import {module}")
    translate_dict = {type(None): "None"}
    global_state = {typename: "a", "typing": typing}
    add_classes(itertools.chain(types.values(), base_classes), translate_dict, global_state)

    del global_state[typename]

    signature = ", ".join(
        (
            f"{name_}: {translate_dict[type_]} = {pprint.pformat(defaults_dict[name_])}"
            if name_ in defaults_dict
            else f"{name_}: {translate_dict[type_]}"
        )
        for name_, type_ in types.items()
    )

    if readonly:
        slots = tuple(f"_{x}" for x in field_names)
        field_definitions = "\n".join(_field_template.format(name=name) for name in field_names)

    else:
        slots = tuple(field_names)
        field_definitions = ""
    init_sig = [f"self.{f_name} = {v_name}" for f_name, v_name in zip(slots, type_dict.keys())]
    tuple_list = [f"self.{name_}" for name_ in slots]
    init_content = "\n        ".join(init_sig) + "\n        self.__post_init__()"
    class_definition = _class_template.format(
        imports="\n".join(import_set),
        typename=typename,
        init_content=init_content,
        tuple_fields=", ".join(tuple_list),
        signature=signature,
        field_names=tuple(field_names),
        slots=slots,
        num_fields=len(field_names),
        arg_list=repr(tuple(field_names)).replace("'", "")[1:-1],
        repr_fmt=", ".join(_repr_template.format(name=name) for name in field_names),
        field_definitions=field_definitions,
        base_classes=", ".join(translate_dict[x] for x in base_classes),
    )

    global_state["__name__"] = f"serialize_{typename}"
    try:
        # pylint: disable=W0122
        exec(class_definition, global_state)  # nosec  # noqa: S102
    except AttributeError:
        print(class_definition, file=sys.stderr)
        raise
    except NameError:
        for i, el in enumerate(class_definition.split("\n"), 1):
            print(f"{i}: {el}", file=sys.stderr)
        raise

    result = global_state[typename]
    result._source = class_definition
    result._field_defaults = defaults_dict  # pylint: disable=protected-access
    result.__annotations__ = types
    with suppress(AttributeError):
        result.__signature__ = inspect.signature(result)
    result._field_types = collections.OrderedDict(types)  # pylint: disable=protected-access
    return result


class BaseMeta(type):
    def __new__(cls, name, bases, attrs):
        if attrs.get("_root", False):
            return super().__new__(cls, name, bases, attrs)
        warnings.warn(
            "BaseSerializableClass is deprecated, use pydantic.BaseModel instead", FutureWarning, stacklevel=2
        )
        types = attrs.get("__annotations__", {})
        defaults = []
        defaults_dict = {}
        for field_name in types:
            if field_name in attrs:
                default_value = attrs[field_name]
                defaults.append(default_value)
                defaults_dict[field_name] = default_value
            elif defaults:
                raise TypeError(
                    "Non-default namedtuple field {field_name} cannot "
                    "follow default field(s) {default_names}".format(
                        field_name=field_name, default_names=", ".join(defaults_dict.keys())
                    )
                )
        if "__readonly__" in attrs:
            readonly = attrs["__readonly__"]
        else:
            readonly = next(
                (el.__readonly__ for el in bases if hasattr(el, "__readonly__")),
                False,
            )

        if "__old_names__" in attrs:
            old_names = attrs["__old_names__"]
            del attrs["__old_names__"]
        else:
            old_names = ()

        result = _make_class(name, types, defaults_dict, list(bases), readonly)
        module = attrs.get("__module__", None)
        if module is None:
            with suppress(AttributeError, ValueError):
                module = sys._getframe(1).f_globals.get("__name__", "__main__")
        if module is not None:
            result.__module__ = module

        for key in attrs:
            if key in _prohibited:
                if key == "__init__":
                    continue
                raise AttributeError(f"Cannot overwrite NamedTuple attribute {key}")
            if key not in _special and key not in result._fields:
                setattr(result, key, attrs[key])
        if "_reloading" not in attrs or not attrs["_reloading"]:
            base_serialize_register.register_class(result, old_names)
        return result


class BaseSerializableClass(metaclass=BaseMeta):
    _root = True
    __readonly__ = True
    __old_names__ = ()

    def __init__(self, *args, **kwargs):
        """declare interface"""

    def __post_init__(self):
        """declare interface"""

    def asdict(self) -> collections.OrderedDict:
        """declare interface"""

    def replace_(self, **_kwargs):
        return self

    def as_tuple(self) -> tuple:
        """declare interface"""

    @classmethod
    def make_(cls, iterable):
        """declare interface"""


class SerializeClassEncoder(json.JSONEncoder):
    # pylint: disable=E0202
    def default(self, o):
        if isinstance(o, pathlib.Path):
            return str(o)
        if isinstance(o, Enum):
            return {"__Enum__": True, "__subtype__": extract_type_info(o.__class__)[0], "value": o.value}
        if isinstance(o, BaseSerializableClass):
            return {"__Serializable__": True, "__subtype__": extract_type_info(o.__class__)[0], **o.asdict()}
        return super().default(o)


def serialize_hook(dkt: dict):
    if "__ReadOnly__" in dkt or "__Serializable__" in dkt:
        # Backward compatibility"
        try:
            cls = base_serialize_register.get_class(dkt["__subtype__"])
        except ValueError:
            dkt["__error__"] = True
            return dkt
        del dkt["__subtype__"]
        dkt.pop("__Serializable__", None)
        dkt.pop("__ReadOnly__", None)
        if isinstance(cls, collections.abc.Iterator):
            keys = set(dkt.keys())
            for el in cls:
                el_keys = set(inspect.signature(el).parameters.keys())
                if keys == el_keys:
                    cls = el
                    break
            else:
                raise ValueError(f"cannot decode {dkt}")
        return cls(**dkt)
    if "__Enum__" in dkt:
        try:
            cls = enum_register.get_class(dkt["__subtype__"])
        except ValueError:
            dkt["__error__"] = True
            return dkt
        del dkt["__Enum__"]
        del dkt["__subtype__"]
        if isinstance(cls, collections.abc.Iterator):
            raise ValueError("Two enum with same name")
        return cls(dkt["value"])
    return dkt



import contextlib
from typing import Optional

from qtpy.QtCore import QObject, QSignalBlocker, Slot
from qtpy.QtGui import QResizeEvent
from qtpy.QtWidgets import QCheckBox, QComboBox, QDoubleSpinBox, QLabel

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.napari_image_view import ImageView
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


class ResultImageView(ImageView):
    """
    :type _settings PartSettings:
    """

    def __init__(self, settings: PartSettings, channel_property: ChannelProperty, name: str):
        super().__init__(settings, channel_property, name)
        self._channel_control_top = True
        self.only_border = QCheckBox("")
        self._set_border_from_settings()
        self.only_border.stateChanged.connect(self._set_border)
        self.opacity = QDoubleSpinBox()
        self.opacity.setRange(0, 1)
        self.opacity.setSingleStep(0.1)
        self._set_opacity_from_settings()
        self.opacity.valueChanged.connect(self._set_opacity)
        self.opacity.setMinimumWidth(500)
        self.channel_control_index = self.btn_layout.indexOf(self.channel_control)
        self.label1 = QLabel("Borders:")
        self.label2 = QLabel("Opacity:")
        self.roi_alternative_select = QComboBox()
        self.roi_alternative_select.currentTextChanged.connect(self._set_roi_alternative_version)
        self.stretch = None

        self.btn_layout.insertWidget(self.channel_control_index + 1, self.label1)
        self.btn_layout.insertWidget(self.channel_control_index + 2, self.only_border)
        self.btn_layout.insertWidget(self.channel_control_index + 3, self.label2)
        self.btn_layout.insertWidget(self.channel_control_index + 4, self.opacity)
        self.btn_layout.insertWidget(self.channel_control_index + 1, self.roi_alternative_select)
        self.channel_control_index = self.btn_layout.indexOf(self.channel_control)

        self.label1.setVisible(False)
        self.label2.setVisible(False)
        self.opacity.setVisible(False)
        self.only_border.setVisible(False)
        self.roi_alternative_select.setVisible(False)

        self.settings.connect_to_profile(f"{self.name}.image_state.only_border", self._set_border_from_settings)
        self.settings.connect_to_profile(f"{self.name}.image_state.opacity", self._set_opacity_from_settings)

    def _set_border(self, value: bool):
        self.settings.set_in_profile(f"{self.name}.image_state.only_border", value)

    def _set_opacity(self, value: float):
        self.settings.set_in_profile(f"{self.name}.image_state.opacity", value)

    def _set_border_from_settings(self):
        self.only_border.setChecked(self.settings.get_from_profile(f"{self.name}.image_state.only_border", False))

    def _set_opacity_from_settings(self):
        self.opacity.setValue(self.settings.get_from_profile(f"{self.name}.image_state.opacity", 1))

    def _set_roi_alternative_version(self, name: str):
        self.roi_alternative_selection = name
        self.update_roi_border()

    def any_roi(self):
        return any(image_info.roi is not None for image_info in self.image_info.values())

    def available_alternatives(self):
        available_alternatives = set()
        for image_info in self.image_info.values():
            if image_info.roi_info.alternative:
                available_alternatives.update(image_info.roi_info.alternative.keys())
        return available_alternatives

    @Slot()
    @Slot(ROIInfo)
    def set_roi(self, roi_info: Optional[ROIInfo] = None, image: Optional[Image] = None) -> None:
        super().set_roi(roi_info, image)
        show = self.any_roi()
        self.label1.setVisible(show)
        self.label2.setVisible(show)
        self.opacity.setVisible(show)
        self.only_border.setVisible(show)
        self.update_alternatives()

    def update_alternatives(self):
        alternatives = self.available_alternatives()
        self.roi_alternative_select.setVisible(bool(alternatives))
        text = self.roi_alternative_select.currentText()
        block = self.roi_alternative_select.signalsBlocked()
        self.roi_alternative_select.blockSignals(True)
        self.roi_alternative_select.clear()
        values = ["ROI", *list(alternatives)]
        self.roi_alternative_select.addItems(values)
        with contextlib.suppress(ValueError):
            self.roi_alternative_select.setCurrentIndex(values.index(text))
        self.roi_alternative_select.blockSignals(block)

    def resizeEvent(self, event: QResizeEvent):
        if event.size().width() > 800 and not self._channel_control_top:
            w = self.btn_layout2.takeAt(0).widget()
            channel_control_index = self.btn_layout.indexOf(self.search_roi_btn) + 1
            self.btn_layout.takeAt(channel_control_index)
            # noinspection PyArgumentList
            self.btn_layout.insertWidget(channel_control_index, w)
            select = self.btn_layout2.takeAt(self.btn_layout2.indexOf(self.roi_alternative_select)).widget()
            self.btn_layout.insertWidget(channel_control_index + 1, select)
            self._channel_control_top = True
        elif event.size().width() <= 800 and self._channel_control_top:
            channel_control_index = self.btn_layout.indexOf(self.channel_control)
            w = self.btn_layout.takeAt(channel_control_index).widget()
            self.btn_layout.insertStretch(channel_control_index, 1)
            # noinspection PyArgumentList
            self.btn_layout2.insertWidget(0, w)
            select = self.btn_layout.takeAt(self.btn_layout.indexOf(self.roi_alternative_select)).widget()
            self.btn_layout2.insertWidget(1, select)
            self._channel_control_top = False


class CompareImageView(ResultImageView):
    def __init__(self, settings: PartSettings, channel_property: ChannelProperty, name: str):
        super().__init__(settings, channel_property, name)
        settings.roi_changed.disconnect(self.set_roi)
        settings.roi_clean.disconnect(self.set_roi)
        settings.compare_segmentation_change.connect(self.set_roi)


class SynchronizeView(QObject):
    def __init__(self, image_view1: ImageView, image_view2: ImageView, parent=None):
        super().__init__(parent)
        self.image_view1 = image_view1
        self.image_view2 = image_view2
        self.synchronize = False
        self.image_view1.view_changed.connect(self.synchronize_views)
        self.image_view2.view_changed.connect(self.synchronize_views)

    def set_synchronize(self, val: bool):
        self.synchronize = val

    @Slot()
    def synchronize_views(self):
        if not self.synchronize or self.image_view1.isHidden() or self.image_view2.isHidden():
            return
        sender = self.sender()
        if sender == self.image_view1:
            origin, dest = self.image_view1, self.image_view2
        else:
            origin, dest = self.image_view2, self.image_view1
        _block = QSignalBlocker(dest)
        if origin.viewer.dims.ndim == dest.viewer.dims.ndim:
            dest.set_state(origin.get_state())



import sys
from abc import ABC
from collections.abc import Iterable
from enum import Enum
from typing import Any, ClassVar, ForwardRef, Optional, Union

import numpy as np
from local_migrator import REGISTER, class_to_str, register_class, rename_key
from pydantic import Field, validator
from sympy import Symbol, symbols

from PartSegCore.algorithm_describe_base import (
    AlgorithmDescribeBase,
    AlgorithmDescribeNotFound,
    base_model_to_algorithm_property,
)
from PartSegCore.universal_const import Units
from PartSegCore.utils import BaseModel
from PartSegImage import Channel
from PartSegImage.image import Spacing


@register_class(
    old_paths=[
        "PartSeg.utils.analysis.statistics_calculation.PerComponent",
        "PartSeg.utils.analysis.measurement_base.PerComponent",
        "segmentation_analysis.statistics_calculation.PerComponent",
    ]
)
class PerComponent(Enum):
    """How measurement should be calculated"""

    No = 1
    Yes = 2
    Mean = 3
    Per_Mask_component = 4

    def __str__(self):
        return self.name.replace("_", " ")


@register_class(
    old_paths=[
        "PartSeg.utils.analysis.statistics_calculation.AreaType",
        "PartSeg.utils.analysis.measurement_base.AreaType",
        "segmentation_analysis.statistics_calculation.AreaType",
    ]
)
class AreaType(Enum):
    """On which area type measurement should be calculated"""

    ROI = 1
    Mask = 2
    Mask_without_ROI = 3

    def __str__(self):
        return self.name.replace("_", " ")


def has_mask_components(component_and_mask_info: Iterable[tuple[PerComponent, AreaType]]) -> bool:
    """Check if any measurement will return value per mask component"""
    return any(
        (cmp == PerComponent.Yes and area != AreaType.ROI) or cmp == PerComponent.Per_Mask_component
        for cmp, area in component_and_mask_info
    )


def has_roi_components(component_and_mask_info: Iterable[tuple[PerComponent, AreaType]]) -> bool:
    """Check if any measurement will return value per ROI component"""
    return any((cmp == PerComponent.Yes and area == AreaType.ROI) for cmp, area in component_and_mask_info)


def _migrate_leaf_dict(dkt):
    from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT

    new_dkt = dkt.copy()
    new_dkt["parameter_dict"] = new_dkt.pop("dict")
    new_dkt["name"] = MEASUREMENT_DICT[new_dkt["name"]].get_name()

    return new_dkt


@register_class(
    version="0.0.2",
    old_paths=[
        "PartSeg.utils.analysis.statistics_calculation.Leaf",
        "PartSeg.utils.analysis.measurement_base.Leaf",
        "segmentation_analysis.statistics_calculation.Leaf",
    ],
    migrations=[("0.0.1", _migrate_leaf_dict), ("0.0.2", rename_key("parameter_dict", "parameters"))],
)
class Leaf(BaseModel):
    """
    Class for describe calculation of basic measurement
    """

    name: str
    parameters: Any = Field(default_factory=dict)
    power: float = 1.0
    area: Optional[AreaType] = None
    per_component: Optional[PerComponent] = None
    channel: Optional[Channel] = None

    @validator("parameters")
    def _validate_parameters(cls, v, values):  # pylint: disable=no-self-use
        if not isinstance(v, dict) or "name" not in values:
            return v
        from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT

        if values["name"] not in MEASUREMENT_DICT:
            return v

        method = MEASUREMENT_DICT[values["name"]]
        if not method.__new_style__ or not method.__argument_class__.__fields__:
            return v

        v = REGISTER.migrate_data(class_to_str(method.__argument_class__), {}, v)
        return method.__argument_class__(**v)

    @validator("per_component")
    def _validate_per_component(cls, v, values):  # pylint: disable=no-self-use
        if not isinstance(v, PerComponent) or "area" not in values or values["area"] is None:
            return v
        if v == PerComponent.Per_Mask_component and values["area"] != AreaType.ROI:
            raise ValueError("Per_Mask_component can be used only with ROI area")
        return v

    def get_channel_num(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> set[Channel]:
        """
        Get set with number of channels needed for calculate this measurement

        :param measurement_dict: dict with all measurement methods.
        :return: set of channels num
        """
        resp = set()
        if self.channel is not None and self.channel.value != -1:
            resp.add(self.channel)
        try:
            measurement_method = measurement_dict[self.name]
            if measurement_method.__new_style__:
                fields = base_model_to_algorithm_property(measurement_method.__argument_class__)
            else:
                fields = measurement_method.get_fields()
            for el in fields:
                if isinstance(el, str):
                    continue
                if el.value_type is Channel:
                    if isinstance(self.parameters, dict):
                        if el.name in self.parameters:
                            resp.add(Channel(self.parameters[el.name]))
                    elif hasattr(self.parameters, el.name):
                        resp.add(getattr(self.parameters, el.name))
        except KeyError as e:
            raise AlgorithmDescribeNotFound(self.name) from e
        return resp

    def _parameters_string(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> str:
        parameters = dict(self.parameters)
        if not parameters and self.channel is None:
            return ""
        arr = []
        if self.channel is not None and self.channel.value != -1:
            arr.append(f"channel={self.channel}")
        if self.name in measurement_dict:
            measurement_method = measurement_dict[self.name]
            fields_dict = measurement_method.get_fields_dict()
            arr.extend(f"{fields_dict[k].user_name}={v}" for k, v in parameters.items())
        else:
            arr.extend(f"{k.replace('_', ' ')}={v}" for k, v in parameters.items())
        return "[" + ", ".join(arr) + "]"

    def _plugin_info(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> str:
        if self.name not in measurement_dict:
            return ""
        measurement_method = measurement_dict[self.name]
        if (
            hasattr(measurement_method, "__module__")
            and measurement_method.__module__.split(".", 1)[0] != "PartSegCore"
        ):
            if getattr(sys, "frozen", False):
                return f"[{measurement_method.__module__.split('.', 2)[1]}] "
            return f"[{measurement_method.__module__.split('.', 1)[0]}] "
        return ""

    def pretty_print(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> str:
        """
        Pretty print for presentation in user interface.

        :param measurement_dict: dict with additional information used for more detailed description
        :return: string with indentation
        """

        resp = self.name
        if self.area is not None:
            resp = f"{self.area} {resp}"
        resp = self._plugin_info(measurement_dict) + resp
        if self.per_component is not None:
            if self.per_component == PerComponent.Yes:
                resp += " per component "
            elif self.per_component == PerComponent.Per_Mask_component:
                resp += " per mask component "
            elif self.per_component == PerComponent.Mean:
                resp += " mean component "
        resp += self._parameters_string(measurement_dict)
        if self.power != 1.0:
            resp += f" to the power {self.power}"
        return resp

    def __str__(self):  # pragma: no cover
        return self.pretty_print({})

    def get_unit(self, ndim: int) -> Symbol:
        """
        Return unit of selected measurement reflecting dimensionality.

        :param ndim: data dimensionality
        """
        from PartSegCore.analysis import MEASUREMENT_DICT

        method = MEASUREMENT_DICT[self.name]
        if self.power != 1:
            return method.get_units(ndim) ** self.power
        return method.get_units(ndim)

    def is_per_component(self) -> bool:
        """If measurement return list of result or single value."""
        return self.per_component in {PerComponent.Yes, PerComponent.Per_Mask_component}

    def need_mask(self) -> bool:
        """If this measurement need mast for proper calculation."""
        return (
            self.area in {AreaType.Mask, AreaType.Mask_without_ROI}
            or self.per_component is PerComponent.Per_Mask_component
        )


def replace(self, **kwargs) -> Leaf:
    for key in list(kwargs.keys()):
        if key == "power":
            continue
        if not hasattr(self, key):
            raise ValueError(f"Unknown parameter {key}")
        if getattr(self, key) is not None and (key != "parameters" or dict(self.parameters)):
            del kwargs[key]

    return self.copy(update=kwargs)


Leaf.replace_ = replace

Node = ForwardRef("Node")


@register_class(
    old_paths=[
        "PartSeg.utils.analysis.statistics_calculation.Node",
        "PartSeg.utils.analysis.measurement_base.Node",
        "segmentation_analysis.statistics_calculation.Node",
    ]
)
class Node(BaseModel):
    """
    Class for describe operation between two measurements
    """

    left: Union[Node, Leaf]
    op: str = Field(
        description="Operation to perform between left and right child. Currently only division (`/`) supported"
    )
    right: Union[Node, Leaf]

    def get_channel_num(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> set[Channel]:
        return self.left.get_channel_num(measurement_dict) | self.right.get_channel_num(measurement_dict)

    def __str__(self):  # pragma: no cover
        left_text = f"({self.left!s})" if isinstance(self.left, Node) else str(self.left)

        right_text = f"({self.right!s})" if isinstance(self.right, Node) else str(self.right)

        return left_text + self.op + right_text

    def pretty_print(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> str:  # pragma: no cover
        left_text = (
            f"({self.left.pretty_print(measurement_dict)})"
            if isinstance(self.left, Node)
            else self.left.pretty_print(measurement_dict)
        )

        right_text = (
            f"({self.right.pretty_print(measurement_dict)})"
            if isinstance(self.right, Node)
            else self.right.pretty_print(measurement_dict)
        )

        return left_text + self.op + right_text

    def get_unit(self, ndim) -> Symbol:
        if self.op == "/":
            return self.left.get_unit(ndim) / self.right.get_unit(ndim)
        raise ValueError(f"Unknown operator '{self.op}'")

    def is_per_component(self) -> bool:
        return self.left.is_per_component() or self.right.is_per_component()

    def need_mask(self):
        return self.left.need_mask() or self.right.need_mask()


Node.update_forward_refs()


@register_class(
    old_paths=[
        "PartSeg.utils.analysis.statistics_calculation.StatisticEntry",
        "PartSeg.utils.analysis.measurement_base.StatisticEntry",
        "segmentation_analysis.statistics_calculation.StatisticEntry",
    ]
)
class MeasurementEntry(BaseModel):
    """Describe single measurement in measurement set"""

    name: str
    calculation_tree: Union[Node, Leaf]

    def get_unit(self, unit: Units, ndim) -> str:
        return str(self.calculation_tree.get_unit(ndim)).format(str(unit))

    def get_channel_num(self, measurement_dict: dict[str, "MeasurementMethodBase"]) -> set[Channel]:
        return self.calculation_tree.get_channel_num(measurement_dict)


class MeasurementMethodBase(AlgorithmDescribeBase, ABC):
    """
    This is base class For all measurement calculation classes
    based on text_info[0] the measurement name will be generated,
    based_on text_info[1] the description is generated
    """

    __argument_class__ = BaseModel

    text_info = "", ""

    need_class_method: ClassVar[list[str]] = [
        "get_description",
        "is_component",
        "calculate_property",
        "get_starting_leaf",
        "get_units",
        "need_channel",
    ]

    @classmethod
    def get_name(cls) -> str:
        """Name of measurement"""
        return str(cls.get_starting_leaf().name)

    @classmethod
    def get_description(cls) -> str:
        """Measurement long description"""
        return "" if isinstance(cls.text_info, str) else cls.text_info[1]

    @classmethod
    def is_component(cls) -> bool:
        """Return information if Need information about components"""
        return False

    @staticmethod
    def calculate_property(
        # image: Image,
        channel: np.ndarray,
        roi: np.ndarray,
        mask: np.ndarray,
        voxel_size: Spacing,
        result_scalar: float,
        roi_alternative: dict[str, np.ndarray],
        roi_annotation: dict[int, Any],
        **kwargs,
    ):
        """
        Main function for calculating measurement

        :param channel: main channel selected for measurement
        :param roi: array representing roi
        :param mask: array representing mask (upper level roi)
        :param voxel_size: size of single voxel in meters
        :param result_scalar: scalar to get proper units in result
        :param roi_alternative: dict with alternative roi representation (for plugin specific mapping)
        :param roi_annotation: dict with roi annotations (for plugin specific mapping)

        List incomplete.
        """
        raise NotImplementedError

    @classmethod
    def get_starting_leaf(cls) -> Leaf:
        """This leaf is put on a default list"""
        if (
            hasattr(cls, "__argument_class__")
            and cls.__argument_class__ is not None
            and cls.__argument_class__ is not BaseModel
        ):
            return Leaf(name=cls._display_name(), parameters=cls.__argument_class__())
        return Leaf(name=cls._display_name())

    @classmethod
    def _display_name(cls):
        return cls.text_info if isinstance(cls.text_info, str) else cls.text_info[0]

    @classmethod
    def get_units(cls, ndim) -> symbols:
        """Return units for measurement. They are shown to user"""
        raise NotImplementedError

    @classmethod
    def need_channel(cls):
        """if need image data"""
        return False

    @staticmethod
    def area_type(area: AreaType):
        """Map chosen area type to proper area type. Allow to correct Area type."""
        return area

    @staticmethod
    def need_full_data():
        return False
